import testing.postgresql
import numpy as np
import asyncpg
import datetime
import itertools
import tempfile
import shutil
import os
import sys
import unittest
import asyncio

from joule.models import DataStream, Element, pipes
from joule.models.data_store import psql_helpers
from joule.models.data_store.timescale import TimescaleStore
from tests import helpers
from tests.models.pipes.reader import QueueReader

SQL_DIR = os.path.join(os.path.dirname(__file__), '..', '..', '..', 'src', 'joule', 'sql')


class TestTimescale(unittest.IsolatedAsyncioTestCase):
    use_default_loop = False
    forbid_get_event_loop = True

    async def asyncSetUp(self):
        # set up the pscql database
        self.psql_dir = tempfile.TemporaryDirectory()
        # run this asynchronously to avoid blocking the event loop
        def start_postgresql_instance():
            return testing.postgresql.Postgresql(base_dir=self.psql_dir.name)
        self.postgresql = await asyncio.to_thread(start_postgresql_instance)
        await asyncio.to_thread(self.postgresql.stop)

        # now that the directory structure is created, customize the *.conf file
        src = os.path.join(os.path.dirname(__file__), "postgresql.conf")
        dest = os.path.join(self.psql_dir.name, "data", "postgresql.conf")
        shutil.copyfile(src, dest)
        # restart the database
        self.postgresql = await asyncio.to_thread(start_postgresql_instance)

        self.db_url = self.postgresql.url()
        conn: asyncpg.Connection = await asyncpg.connect(self.db_url)
        await conn.execute("CREATE EXTENSION IF NOT EXISTS timescaledb CASCADE ")
        await conn.execute("CREATE USER joule_module")
        # load custom functions
        for file in os.listdir(SQL_DIR):
            file_path = os.path.join(SQL_DIR, file)
            with open(file_path, 'r') as f:
                await conn.execute(f.read())
        await conn.close()

    async def asyncTearDown(self):
        await asyncio.to_thread(self.postgresql.stop)
        self.psql_dir.cleanup()

    async def test_runner(self):
        tests = [
            self._test_merge_gaps_on_write,
            self._test_basic_insert_extract,
            self._test_extract_data_with_intervals,
            self._test_extract_decimated_data,
            self._test_db_info,
            self._test_info,
            self._test_intervals,
            self._test_remove,
            self._test_destroy,
            self._test_row_count,
            self._test_actions_on_empty_streams,
            self._test_consolidate,
            self._test_consolidate_with_time_bounds,
            self._test_remove_decimations,
            self._test_redecimates_data]
        for test in tests:
            conn: asyncpg.Connection = await asyncpg.connect(self.db_url)
            await conn.execute("DROP SCHEMA IF EXISTS data CASCADE")
            await conn.execute("CREATE SCHEMA data")
            await conn.execute("GRANT ALL ON SCHEMA data TO public")

            self.store = TimescaleStore(self.db_url, 0, 60,)
            await self.store.initialize([])
            # make a sample stream with data
            self.test_stream = DataStream(id=100, name="stream1", datatype=DataStream.DATATYPE.FLOAT32,
                                          keep_us=DataStream.KEEP_ALL, decimate=True,
                                          elements=[Element(name="e%d" % x) for x in range(3)])
            pipe = pipes.LocalPipe(self.test_stream.layout)
            self.test_data = helpers.create_data(layout=self.test_stream.layout, length=1005)
            task = self.store.spawn_inserter(self.test_stream, pipe)
            await pipe.write(self.test_data)
            await pipe.close()
            runner = await task
            await runner
            await conn.close()
            # await self.store.initialize([])
            await test()
            # simulate the nose2 test output
            sys.stdout.write('o')
            await self.store.close()
            sys.stdout.flush()

    async def _test_basic_insert_extract(self):
        stream_id = 990
        self.store.extract_block_size = 500
        psql_types = ['double precision', 'real', 'bigint', 'integer', 'smallint']
        datatypes = [DataStream.DATATYPE.FLOAT64, DataStream.DATATYPE.FLOAT32, DataStream.DATATYPE.INT64,
                     DataStream.DATATYPE.INT32, DataStream.DATATYPE.INT16]
        conn: asyncpg.Connection = await asyncpg.connect(self.db_url)
        for i in range(len(datatypes)):
            datatype = datatypes[i]
            psql_type = psql_types[i]
            for n_elements in range(1, 5):
                test_stream = DataStream(id=stream_id, name="stream1", datatype=datatype, keep_us=DataStream.KEEP_ALL,
                                         elements=[Element(name="e%d" % x) for x in range(n_elements)])
                test_stream.decimate = True
                source = QueueReader()
                pipe = pipes.InputPipe(stream=test_stream, reader=source)
                nrows = 955
                data = helpers.create_data(layout=test_stream.layout, length=nrows)
                task = await self.store.spawn_inserter(test_stream, pipe)
                for chunk in helpers.to_chunks(data, 300):
                    await source.put(chunk.tobytes())
                await task

                # make sure the correct tables have been created
                records = await conn.fetch('''SELECT table_name FROM information_schema.tables 
                                                    WHERE table_schema='data';''')
                tables = list(itertools.chain(*records))
                for table in ['stream%d' % stream_id, 'stream%d_intervals' % stream_id]:
                    self.assertIn(table, tables)

                # check the column data types
                records = await conn.fetch('''SELECT column_name, data_type FROM information_schema.columns 
                                                            WHERE table_name='stream%d' AND table_schema='data';''' % stream_id)
                (names, types) = zip(*records)
                expected_elements = ['time'] + ['elem%d' % x for x in range(n_elements)]
                self.assertCountEqual(names, expected_elements)
                expected_psql_types = tuple(['timestamp without time zone'] + [psql_type for x in range(n_elements)])
                self.assertEqual(types, expected_psql_types)
                self.assertEqual(len(records), n_elements + 1)

                # extract raw data
                extracted_data = []

                async def callback(rx_data, layout, factor):
                    self.assertEqual(layout, test_stream.layout)
                    self.assertEqual(factor, 1)
                    extracted_data.append(rx_data)

                await self.store.extract(test_stream, start=None, end=None, callback=callback)
                extracted_data = np.hstack(extracted_data)
                np.testing.assert_array_equal(extracted_data, data)

                level = 64
                data_mean = np.mean(extracted_data['data'][:level], axis=0)
                data_max = np.max(extracted_data['data'][:level], axis=0)
                data_min = np.min(extracted_data['data'][:level], axis=0)

                # extract decimated data
                async def d_callback(rx_data, layout, factor):
                    self.assertEqual(layout, test_stream.decimated_layout)
                    self.assertEqual(factor, level)
                    extracted_data.append(rx_data)

                extracted_data = []
                await self.store.extract(test_stream, decimation_level=level,
                                         start=None, end=None, callback=d_callback)
                extracted_data = np.hstack(extracted_data)
                expected = np.hstack((data_mean, data_min, data_max))
                np.testing.assert_array_almost_equal(expected, extracted_data['data'][0])
                stream_id += 1
        await conn.close()

    async def _test_extract_data_with_intervals(self):
        test_stream = DataStream(id=1, name="stream1", datatype=DataStream.DATATYPE.FLOAT32, keep_us=DataStream.KEEP_ALL,
                                 decimate=True, elements=[Element(name="e%d" % x) for x in range(3)])
        pipe = pipes.LocalPipe(test_stream.layout)
        nrows = 955
        data = helpers.create_data(layout=test_stream.layout, length=nrows)
        task = await self.store.spawn_inserter(test_stream, pipe)
        for chunk in helpers.to_chunks(data, 300):
            await pipe.write(chunk)
            await pipe.close_interval()
        await pipe.close()
        await task

        # extract data
        extracted_data = []

        async def callback(rx_data, layout, factor):
            self.assertEqual(layout, test_stream.layout)
            self.assertEqual(factor, 1)
            extracted_data.append(rx_data)

        await self.store.extract(test_stream, start=None, end=None, callback=callback)
        extracted_data = np.hstack(extracted_data)
        # check for interval boundaries
        np.testing.assert_array_equal(extracted_data[300], pipes.interval_token(test_stream.layout))
        np.testing.assert_array_equal(extracted_data[601], pipes.interval_token(test_stream.layout))
        np.testing.assert_array_equal(extracted_data[902], pipes.interval_token(test_stream.layout))

    async def _test_merge_gaps_on_write(self):
        # make sure successive writes less than a merge_gap apart are merged
        test_stream = DataStream(id=1, name="stream1", datatype=DataStream.DATATYPE.FLOAT32, keep_us=DataStream.KEEP_ALL,
                                 decimate=True, elements=[Element(name="e%d" % x) for x in range(3)])
        nrows = 300
        data = helpers.create_data(layout=test_stream.layout, length=nrows, step=1000, start=15e6)
        data_1 = data[:100]
        data_2 = data[100:200]
        data_3 = data[200:]
        # write the first chunk
        pipe = pipes.LocalPipe(test_stream.layout)
        task = await self.store.spawn_inserter(test_stream, pipe)
        await pipe.write(data_1)
        await pipe.close()
        await task

        # write the second chunk inside the merge_gap setting
        pipe = pipes.LocalPipe(test_stream.layout)
        task = await self.store.spawn_inserter(test_stream, pipe, merge_gap=1100)
        await pipe.write(data_2)
        await pipe.close()
        await task

        # write the third chunk outside the merge_gap setting
        pipe = pipes.LocalPipe(test_stream.layout)
        task = await self.store.spawn_inserter(test_stream, pipe, merge_gap=900)
        await pipe.write(data_3)
        await pipe.close()
        await task

        # extract data
        extracted_data = []

        intervals = await self.store.intervals(test_stream,start=None,end=None)
        #print(intervals)
        async def callback(rx_data, layout, factor):
            self.assertEqual(layout, test_stream.layout)
            self.assertEqual(factor, 1)
            extracted_data.append(rx_data)

        await self.store.extract(test_stream, start=None, end=None, callback=callback)
        extracted_data = np.hstack(extracted_data)
        self.assertEqual(len(extracted_data),301)
        # check for interval boundaries
        for i in range(200):
            np.testing.assert_array_equal(extracted_data[i], data[i])
        np.testing.assert_array_equal(extracted_data[200], pipes.interval_token(test_stream.layout))
        for i in range(201, 301):
            np.testing.assert_array_equal(extracted_data[i], data[i-1])
       

    async def _test_extract_decimated_data(self):

        extracted_data = []

        async def callback(rx_data, layout, factor):
            self.assertEqual(layout, self.test_stream.decimated_layout)
            self.assertEqual(factor, 4)
            extracted_data.append(rx_data)

        await self.store.extract(self.test_stream, max_rows=16,
                                 start=self.test_data['timestamp'][66],
                                 end=self.test_data['timestamp'][64 + 4 * 16],
                                 callback=callback)
        extracted_data = np.hstack(extracted_data)
        self.assertLessEqual(len(extracted_data), 16)
        self.assertGreater(len(extracted_data), 4)

    async def _test_remove_decimations(self):

        # make sure the stream is decimated:
        conn: asyncpg.Connection = await asyncpg.connect(self.db_url)
        table_names = await psql_helpers.get_decimation_table_names(conn, self.test_stream, with_schema=False)
        for table in [f'stream100_{4**i}' for i in range(1,6)]:
           self.assertIn(table, table_names)
        await self.store.drop_decimations(self.test_stream)
        # drop the decimations
        table_names = await psql_helpers.get_decimation_table_names(conn, self.test_stream, with_schema=False)
        # make sure the tables are dropped
        self.assertEqual(len(table_names), 0)
        await conn.close()

    async def _test_redecimates_data(self):
        # create decimated data
        test_stream = DataStream(id=1, name="stream1", datatype=DataStream.DATATYPE.FLOAT32,
                                 keep_us=DataStream.KEEP_ALL,
                                 elements=[Element(name="e%d" % x) for x in range(3)])
        test_stream.decimate = True
        source = QueueReader()
        pipe = pipes.InputPipe(stream=test_stream, reader=source)
        nrows = 1024
        data = helpers.create_data(layout=test_stream.layout, length=nrows)
        task = await self.store.spawn_inserter(test_stream, pipe)
        for chunk in helpers.to_chunks(data, 16):
            await source.put(chunk.tobytes())
            await source.put(pipes.interval_token(test_stream.layout).tobytes())
        await task # put the data in the database, 16 rows per interval
        conn: asyncpg.Connection = await asyncpg.connect(self.db_url)
        table_names = await psql_helpers.get_decimation_table_names(conn, test_stream, with_schema=False)
        # interval breaks cause decimation to stop after level 16 (level 64 is created but empty)
        self.assertEqual(len(table_names),3)
        orig_data = await conn.fetch("SELECT * FROM data.stream1_16")
        # redecimate doesn't change anything
        await self.store.decimate(test_stream)
        redecimated_data = await conn.fetch("SELECT * FROM data.stream1_16")
        self.assertEqual(orig_data, redecimated_data)

        intervals = await self.store.intervals(test_stream, start=None, end=None)
        self.assertEqual(len(intervals), 64)
        # now consolidate the intervals
        await self.store.consolidate(test_stream, start=None, end=None,
                                     max_gap = data['timestamp'][1]-data['timestamp'][0])
        intervals = await self.store.intervals(test_stream, start=None, end=None)
        self.assertEqual(len(intervals), 1)
        # redecimate
        await self.store.decimate(test_stream)
        table_names = await psql_helpers.get_decimation_table_names(conn, test_stream, with_schema=False)
        for table in [f'stream1_{4**i}' for i in range(1,6)]:
           self.assertIn(table, table_names)
        await conn.close()

    async def _test_nondecimating_inserter(self):
        # TODO
        pass

    async def _test_row_count(self):
        test_data = helpers.create_data(layout=self.test_stream.layout, length=10000)
        test_stream = DataStream(id=95, name="stream1", datatype=DataStream.DATATYPE.FLOAT32, keep_us=DataStream.KEEP_ALL,
                                 decimate=True, elements=[Element(name="e%d" % x) for x in range(3)])
        pipe = pipes.LocalPipe(test_stream.layout)
        task = await self.store.spawn_inserter(test_stream, pipe)
        await pipe.write(test_data)
        await pipe.close()
        await task
        conn: asyncpg.Connection = await asyncpg.connect(self.db_url)
        # test to make sure nrows is within 10% of actual value
        # Test: [start, end]
        nrows = await psql_helpers.get_row_count(conn, test_stream,
                                                 None,
                                                 None)
        self.assertGreater(nrows, len(test_data) * 0.9)
        # Test: [ts,end]
        nrows = await psql_helpers.get_row_count(conn, test_stream, test_data[len(test_data) // 2][0], None)
        self.assertLess(abs(nrows - len(test_data) // 2), 0.1 * len(test_data))
        # Test: [start, ts]
        nrows = await psql_helpers.get_row_count(conn, test_stream,
                                                 None,
                                                 test_data[len(test_data) // 3][0])
        self.assertLess(abs(nrows - len(test_data) // 3), 0.1 * len(test_data))

        # Test: [ts, ts]
        nrows = await psql_helpers.get_row_count(conn, test_stream,
                                                 test_data[2 * len(test_data) // 6][0],
                                                 test_data[3 * len(test_data) // 6][0])
        self.assertLess(abs(nrows - len(test_data) // 6), 0.1 * len(test_data))

        # Test: [ts, ts] (no data)
        nrows = await psql_helpers.get_row_count(conn, test_stream,
                                                 test_data['timestamp'][0] - 100,
                                                 test_data['timestamp'][0] - 50)
        self.assertEqual(0, nrows)

        # Test row count for stream with no data tables
        empty_stream = DataStream(id=96, name="empty", datatype=DataStream.DATATYPE.FLOAT64,
                                  keep_us=100, decimate=True, elements=[Element(name="e%d" % x) for x in range(8)])
        nrows = await psql_helpers.get_row_count(conn, empty_stream,
                                                 None,
                                                 None)
        self.assertEqual(0, nrows)
        nrows = await psql_helpers.get_row_count(conn, empty_stream,
                                                 test_data[len(test_data) // 2][0],
                                                 None)
        self.assertEqual(0, nrows)
        nrows = await psql_helpers.get_row_count(conn, test_stream,
                                                 test_data['timestamp'][0] - 100,
                                                 test_data['timestamp'][0] - 50)
        self.assertEqual(0, nrows)
        await conn.close()

    async def _test_remove(self):

        #  XXXXXXX------XXXXXXX
        #        ^|     ^
        # remove a chunk of data from the middle
        await self.store.remove(self.test_stream,
                                self.test_data['timestamp'][300],
                                self.test_data['timestamp'][400])
        # XXXXXXX------===XXXX
        #       ^|        ^
        # remove another chunk
        await self.store.remove(self.test_stream,
                                self.test_data['timestamp'][350],
                                self.test_data['timestamp'][500])

        # XXX___----===XXX
        #   ^|         ^
        await self.store.remove(self.test_stream,
                                self.test_data['timestamp'][250],
                                self.test_data['timestamp'][300])

        # extract the data, should have an interval gap between 249 and 500
        # and a *single* interval boundary at 250

        extracted_data = []

        async def callback(rx_data, layout, decimated):
            extracted_data.append(rx_data)

        await self.store.extract(self.test_stream, start=None, end=None,
                                 callback=callback)
        extracted_data = np.hstack(extracted_data)

        # beginning is unchanged
        np.testing.assert_array_equal(extracted_data[:249], self.test_data[:249])
        # interval boundary marking missing data
        np.testing.assert_array_equal(extracted_data[250],
                                      pipes.interval_token(self.test_stream.layout))
        # end is unchanged (closing interval boundary ignored)
        np.testing.assert_array_equal(extracted_data[251:], self.test_data[500:])

        # two intervals of data
        intervals = await self.store.intervals(self.test_stream, start=None, end=None)
        ts = self.test_data['timestamp']
        expected = [[ts[0], ts[249] + 1],
                    [ts[500], ts[-1] + 1]]
        self.assertEqual(intervals, expected)

    async def _test_destroy(self):
        await self.store.destroy(self.test_stream)
        records = await self.store.info([self.test_stream])
        info = records[self.test_stream.id]
        self.assertIsNone(info.start)
        self.assertIsNone(info.end)
        self.assertEqual(info.total_time, 0)
        self.assertEqual(info.rows, 0)
        self.assertEqual(info.bytes, 0)

    async def _test_intervals(self):

        ts = self.test_data['timestamp']

        async def _insert_boundary(_conn, _ts: int):
            dt = datetime.datetime.fromtimestamp(_ts / 1e6, tz=datetime.timezone.utc)
            query = "INSERT INTO data.stream100_intervals (time) VALUES ('%s')" % dt
            await _conn.execute(query)

        conn: asyncpg.Connection = await asyncpg.connect(self.db_url)

        # XXXX|XXXXX|XXXXX
        await _insert_boundary(conn, ts[99] + 1)
        await _insert_boundary(conn, ts[199] + 1)
        expected = [[ts[0], ts[99] + 1],
                    [ts[100], ts[199] + 1],
                    [ts[200], ts[-1] + 1]]
        intervals = await self.store.intervals(self.test_stream, None, None)
        self.assertEqual(expected, intervals)

        # insert an extra boundary that does not create a new interval
        # XXXX||XXXXX|XXXXX
        await _insert_boundary(conn, ts[99] + 10)

        intervals = await self.store.intervals(self.test_stream, None, None)
        self.assertEqual(expected, intervals)

        # insert boundaries at the beginning and end of the data,
        # this should also not create any new intervals
        # |XXXX||XXXXX|XXXXX|

        await _insert_boundary(conn, ts[0] - 1)
        await _insert_boundary(conn, ts[-1] + 1)
        intervals = await self.store.intervals(self.test_stream, None, None)
        self.assertEqual(expected, intervals)

        # insert an actual boundary, this should create a new interval
        # |XXXX||XXXXX|XXX|X|
        await _insert_boundary(conn, ts[780] + 1)
        intervals = await self.store.intervals(self.test_stream, None, None)
        expected = [[ts[0], ts[99] + 1],
                    [ts[100], ts[199] + 1],
                    [ts[200], ts[780] + 1],
                    [ts[781], ts[-1] + 1]]
        self.assertEqual(expected, intervals)
        await conn.close()

    async def _test_consolidate(self):
        # intervals less than max_gap us apart are consolidated
        # data: 100 samples spaced at 1000us
        test_stream = DataStream(id=1, name="stream1", datatype=DataStream.DATATYPE.FLOAT32, keep_us=DataStream.KEEP_ALL,
                                 decimate=True, elements=[Element(name="e%d" % x) for x in range(3)])
        nrows = 955
        orig_data = helpers.create_data(layout=test_stream.layout, length=nrows)

        rx_chunks = []
        async def callback(rx_data, layout, factor):
            #if rx_data[0] != pipes.interval_token(layout):
            if rx_data[-1][0]==0:
                rx_chunks.append(rx_data[:-1])
            else:
                rx_chunks.append(rx_data)

        chunks = [orig_data[:300], orig_data[305:400], orig_data[402:700], orig_data[800:]]
        # data: |++++++|  |+++++++++|    |++++++|    |++++|
        #               ^--5000 us    ^--2000 us   ^---0.1 sec (retained)
        pipe = pipes.LocalPipe(test_stream.layout)
        task = await self.store.spawn_inserter(test_stream, pipe)
        for chunk in chunks:
            await pipe.write(chunk)
            await pipe.close_interval()
        await pipe.close()
        await task

        await self.store.consolidate(test_stream, start=None, end=None, max_gap=6e3)
        await self.store.extract(test_stream, start=None, end=None, callback=callback)
        # should only be two intervals left (the first three are consolidated)
        np.testing.assert_array_equal(rx_chunks[0], np.hstack(chunks[:3]))
        np.testing.assert_array_equal(rx_chunks[1], np.hstack(chunks[3:]))
        self.assertEqual(len(rx_chunks), 2)
        await self.store.remove(test_stream, start=None, end=None) # drop the data to get ready for the next test

        chunks = [orig_data[:300], orig_data[305:400], orig_data[402:700], orig_data[800:850], orig_data[852:]]
        # data: |++++++|  |+++++++++|    |++++++|    |++++|  |++++|
        #               ^--5000 us    ^--2000 us   |        ^--- 2000 us
        #                                          `---0.1 sec (retained)
        pipe = pipes.LocalPipe(test_stream.layout)
        task = await self.store.spawn_inserter(test_stream, pipe)
        for chunk in chunks:
            await pipe.write(chunk)
            await pipe.close_interval()
        await pipe.close()
        await task

        rx_chunks = [] # flush the rx buffer for the next test
        await self.store.consolidate(test_stream, start=None, end=None, max_gap=6e3)
        await self.store.extract(test_stream, start=None, end=None, callback=callback)

        # should only be two intervals left (the first two are consolidated)
        np.testing.assert_array_equal(rx_chunks[0], np.hstack(chunks[:3]))
        np.testing.assert_array_equal(rx_chunks[1], np.hstack(chunks[3:]))
        self.assertEqual(len(rx_chunks), 2)

    async def _test_consolidate_with_time_bounds(self):
        # intervals less than max_gap us apart between start and end are consolidated
        # data: 100 samples spaced at 1000us
        test_stream = DataStream(id=1, name="stream1", datatype=DataStream.DATATYPE.FLOAT32, keep_us=DataStream.KEEP_ALL,
                                 decimate=True, elements=[Element(name="e%d" % x) for x in range(3)])
        pipe = pipes.LocalPipe(test_stream.layout)
        nrows = 955
        orig_data = helpers.create_data(layout=test_stream.layout, length=nrows)
        chunks = [orig_data[:300], orig_data[305:400], orig_data[402:700], orig_data[800:850],orig_data[852:]]
        # data: |++++++|  |+++++++++|    |++++++|    |++++|  |++++|
        #               ^--(retained) ^--2000 us   |        ^--- 2000 us (retained)
        #                                          `---0.1 sec (retained)
        task = await self.store.spawn_inserter(test_stream, pipe)
        for chunk in chunks:
            await pipe.write(chunk)
            await pipe.close_interval()
        await pipe.close()
        await task

        # extract data
        rx_chunks = []

        async def callback(rx_data, layout, factor):
            if rx_data[-1][0]==0:
                rx_chunks.append(rx_data[:-1])
            else:
                rx_chunks.append(rx_data)

        await self.store.consolidate(test_stream, start=chunks[1][3]['timestamp'],
                                     end=chunks[3][3]['timestamp'], max_gap=6e3)
        await self.store.extract(test_stream, start=None, end=None, callback=callback)

        # should only be four intervals left (the inner short intervals are consolidated)
        np.testing.assert_array_equal(rx_chunks[0], chunks[0])
        np.testing.assert_array_equal(rx_chunks[1], np.hstack(chunks[1:3]))
        np.testing.assert_array_equal(rx_chunks[2], chunks[3])
        np.testing.assert_array_equal(rx_chunks[3], chunks[4])

        self.assertEqual(len(rx_chunks), 4)

    async def _test_db_info(self):
        db_info = await self.store.dbinfo()
        self.assertTrue(os.path.isdir(db_info.path))
        self.assertGreater(db_info.other, 0)
        self.assertGreater(db_info.reserved, 0)
        self.assertGreater(db_info.free, 0)
        self.assertGreater(db_info.size, 0)

    async def _test_info(self):
        # create another stream
        empty_stream = DataStream(id=103, name="empty stream", datatype=DataStream.DATATYPE.INT32,
                                  keep_us=DataStream.KEEP_ALL, decimate=True,
                                  elements=[Element(name="e%d" % x) for x in range(8)])
        stream2 = DataStream(id=104, name="stream2", datatype=DataStream.DATATYPE.INT32,
                             keep_us=DataStream.KEEP_ALL, decimate=True,
                             elements=[Element(name="e%d" % x) for x in range(8)])
        pipe = pipes.LocalPipe(stream2.layout)
        test_data = helpers.create_data(layout=stream2.layout, length=800)
        task = await self.store.spawn_inserter(stream2, pipe)
        await pipe.write(test_data)
        await pipe.close()
        await task
        records = await self.store.info([self.test_stream, stream2, empty_stream])
        # check stream1
        info = records[self.test_stream.id]
        self.assertEqual(info.start, self.test_data['timestamp'][0])
        self.assertEqual(info.end, self.test_data['timestamp'][-1])
        self.assertEqual(info.total_time, info.end - info.start)
        # rows are approximate
        self.assertLess(abs(len(self.test_data) - info.rows), len(self.test_data) * 0.1)
        self.assertGreater(info.bytes, 0)

        # check stream2
        info = records[stream2.id]
        self.assertEqual(info.start, test_data['timestamp'][0])
        self.assertEqual(info.end, test_data['timestamp'][-1])
        self.assertEqual(info.total_time, info.end - info.start)
        self.assertLess(abs(len(test_data) - info.rows), len(test_data) * 0.1)
        self.assertGreater(info.bytes, 0)

        # check the empty stream
        info = records[empty_stream.id]
        self.assertIsNone(info.start)
        self.assertIsNone(info.end)
        self.assertEqual(info.total_time, 0)
        self.assertEqual(info.rows, 0)
        self.assertEqual(info.bytes, 0)

    async def _test_actions_on_empty_streams(self):
        # make sure all actions can be performed on streams with no data
        empty_stream = DataStream(id=103, name="empty stream", datatype=DataStream.DATATYPE.INT32,
                                  keep_us=DataStream.KEEP_ALL, decimate=True,
                                  elements=[Element(name="e%d" % x) for x in range(8)])

        # ==== extract =====
        cb_executed = False
        cb_layout = None
        cb_factor = None

        async def callback(rx_data, layout, factor):
            nonlocal cb_executed, cb_layout, cb_factor
            cb_executed = True
            self.assertEqual(len(rx_data), 0)
            cb_layout = layout
            cb_factor = factor

        # stream with no data tables
        await self.store.extract(empty_stream, start=None, end=None, callback=callback)
        self.assertTrue(cb_executed)
        self.assertEqual(cb_layout, empty_stream.layout)
        self.assertEqual(cb_factor, 1)
        cb_executed = False
        await self.store.extract(empty_stream, start=100, end=None, callback=callback)
        self.assertTrue(cb_executed)
        self.assertEqual(cb_layout, empty_stream.layout)
        self.assertEqual(cb_factor, 1)
        cb_executed = False
        await self.store.extract(empty_stream, start=100, end=300, callback=callback)
        self.assertTrue(cb_executed)
        self.assertEqual(cb_layout, empty_stream.layout)
        self.assertEqual(cb_factor, 1)
        cb_executed = False
        await self.store.extract(empty_stream, start=None, end=100, callback=callback)
        self.assertTrue(cb_executed)
        self.assertEqual(cb_layout, empty_stream.layout)
        self.assertEqual(cb_factor, 1)
        cb_executed = False
        await self.store.extract(empty_stream, start=100, end=300, max_rows=40, callback=callback)
        self.assertTrue(cb_executed)
        self.assertEqual(cb_layout, empty_stream.layout)
        self.assertEqual(cb_factor, 1)
        cb_executed = False
        await self.store.extract(empty_stream, start=100, end=300, decimation_level=40, callback=callback)
        self.assertTrue(cb_executed)
        self.assertEqual(cb_layout, empty_stream.decimated_layout)
        self.assertEqual(cb_factor, 40)

        # non-existent decimation level
        await self.store.extract(self.test_stream, start=None, end=None, decimation_level=97,
                                 callback=callback)

        #  ==== intervals ====
        intervals = await self.store.intervals(empty_stream, None, None)
        self.assertEqual(0, len(intervals))
        intervals = await self.store.intervals(empty_stream, 100, None)
        self.assertEqual(0, len(intervals))
        intervals = await self.store.intervals(empty_stream, None, 100)
        self.assertEqual(0, len(intervals))
        intervals = await self.store.intervals(empty_stream, 100, 300)
        self.assertEqual(0, len(intervals))

        #  ==== remove ====
        await self.store.remove(empty_stream, None, None)
        await self.store.remove(empty_stream, 100, None)
        await self.store.remove(empty_stream, None, 100)
        await self.store.remove(empty_stream, 100, 300)

        # ==== destroy ====
        await self.store.destroy(empty_stream)



from click.testing import CliRunner
from tests.cli.fake_joule import FakeJoule, FakeJouleTestCase
from joule.cli import main
from ..fake_joule import print_result_on_error


class TestMasterDelete(FakeJouleTestCase):

    def test_removes_followers(self):
        server = FakeJoule()
        self.start_server(server)
        runner = CliRunner()
        result = runner.invoke(main, ['master', 'delete','user','M1'])
        print_result_on_error(result)
        self.assertEqual(result.exit_code, 0)
        output = result.output
        # make sure master is removed
        self.assertIn('M1', output)
        self.stop_server()


    def test_when_server_returns_error_code(self):
        server = FakeJoule()
        error_msg = "test error"
        error_code = 500
        server.response = error_msg
        server.http_code = error_code
        server.stub_master = True
        self.start_server(server)
        runner = CliRunner()
        result = runner.invoke(main, ['master', 'delete','joule','M1'])
        self.assertIn('%d' % error_code, result.output)
        self.assertIn(error_msg, result.output)
        self.assertEqual(result.exit_code, 1)
        self.stop_server()



import argparse
import asyncio
import os
import signal
import sys
import time
import threading
import tempfile
from aiohttp import web
from unittest import mock
import unittest
from joule.client import BaseModule
from tests import helpers
import warnings

warnings.simplefilter('always')


class SimpleModule(BaseModule):

    async def run_as_task(self, parsed_args, app):
        self.completed = False
        return asyncio.create_task(self.stub(parsed_args.stop_on_request))

    # generates warnings because the pipe variable is none
    def routes(self):
        return [web.get('/', self.index)]

    async def index(self, request):
        return web.Response(text="hello world")

    async def stub(self, stop_on_request):
        while True:
            if stop_on_request and self.stop_requested:
                break
            await asyncio.sleep(0.01)
        # only get here if module listens for stop request
        self.completed = True


class NetworkedModule(BaseModule):
    async def run_as_task(self, parsed_args, app):
        return asyncio.create_task(self._build_pipes(parsed_args))


class TestBaseModule(helpers.AsyncTestCase):

    def sigint(self):
        time.sleep(1)
        os.kill(os.getpid(), signal.SIGINT)

    # this test starts a local server because the module
    # is run in a standalone configuration or the data_app flag is false in the config
    def test_stops_on_sigint(self):
        module = SimpleModule()
        # generate a warning for the socket
        args = argparse.Namespace(socket='unset', pipes='unset',
                                  port=None, host='127.0.0.1',
                                  api_socket='dummval', node="",
                                  url="http://localhost:8088",
                                  stop_on_request=True)
        # fire sigint
        t = threading.Thread(target=self.sigint)
        # run the module
        t.start()
        with self.assertLogs(level="INFO") as logs:
            module.start(args)
        all_logs = '\n'.join(logs.output).lower()
        self.assertTrue('starting web server' in all_logs)
        asyncio.set_event_loop(self.loop)
        t.join()
        self.assertTrue(module.completed)


    # this test starts a local server because the module
    # is run in a standalone configuration or the data_app flag is false in the config
    def test_cancels_execution_if_necessary(self):
        module = SimpleModule()
        module.STOP_TIMEOUT = 0.1
        args = argparse.Namespace(socket='unset',
                                  port=None,
                                  host='127.0.0.1',
                                  pipes='unset',
                                  api_socket='dummyval', node="",
                                  url="http://localhost:8088",
                                  stop_on_request=False)
        # fire sigint
        t = threading.Thread(target=self.sigint)
        # run the module
        t.start()
        with self.assertLogs() as logs:
            module.start(args)
        all_logs = '\n'.join(logs.output).lower()
        self.assertTrue('starting web server' in all_logs)
        asyncio.set_event_loop(self.loop)
        t.join()
        self.assertFalse(module.completed)


    # builds network pipes if pipe arg is 'unset'
    def test_builds_network_pipes(self):
        #### THIS TESTS OLD CODE, MIGRATE TO BUILD_PIPES_NEW (still used by composite module) ####
        built_pipes = False

        async def mock_builder(inputs, outputs, streams, node, start_time, end_time, live, force):
            nonlocal built_pipes
            built_pipes = True
            self.assertEqual(len(inputs), 2)
            self.assertEqual(inputs['input1'], '/test/input1:float32[x,y,z]')
            self.assertEqual(len(outputs), 2)
            self.assertEqual(outputs['output2'], '/test/output2:float32[x,y,z]')
            return {}, {}

        module_mock = mock.Mock()
        module_mock.build_network_pipes = mock_builder
        with mock.patch.dict(sys.modules,
                             {'joule.client.helpers.pipes': module_mock}):
            with tempfile.NamedTemporaryFile() as f:
                module = NetworkedModule()
                f.write(str.encode(
                    """
                    [Main]
                    name = test
                    exec_cmd = runit.sh
                    [Arguments]
    
                    [Inputs]
                    input1 = /test/input1:float32[x,y,z]
                    input2 = /test/input2:float32[x,y,z]
    
                    [Outputs]
                    output1 = /test/output1:float32[x,y,z]
                    output2 = /test/output2:float32[x,y,z]
                    """))
                f.flush()
                # set api_socket to dummyval to prevent module from requesting
                # the node from the user configs which is not present here
                args = argparse.Namespace(socket='unset', pipes='unset',
                                          api_socket='dummyval', node="",
                                          start_time='1 hour ago', end_time=None, force=True,
                                          module_config=f.name,
                                          stream_configs='unset',
                                          new=False,
                                          live=False,
                                          url='http://localhost:8088')
                module.start(args)
                self.assertTrue(built_pipes)



import click
from joule.cli.config import pass_config
from joule.api import set_default_node
from joule.errors import ApiError

@click.command(name="default")
@click.argument("name")
@pass_config
def node_default(config, name):
    """Change the default node."""
    try:
        set_default_node(name)
        click.echo("Set [%s] as the default node" % name)
    except ApiError as e:
        raise click.ClickException(str(e))



from click.testing import CliRunner
import os
import warnings
import json
from tests.cli.fake_joule import FakeJoule, FakeJouleTestCase
from joule.models import Master
from joule.cli import main
from ..fake_joule import print_result_on_error



class TestMasterList(FakeJouleTestCase):

    def test_lists_masters(self):
        server = FakeJoule()
        masters = [Master(name="M1", type=Master.TYPE.USER, key="1234"),
                   Master(name="M2", type=Master.TYPE.JOULE_NODE, key="5678"),
                   Master(name="M3", type=Master.TYPE.LUMEN_NODE, key="abcd")]
        server.response = json.dumps([f.to_json() for f in masters])
        self.start_server(server)
        runner = CliRunner()
        result = runner.invoke(main, ['master', 'list'])
        print_result_on_error(result)
        self.assertEqual(result.exit_code, 0)
        output = result.output
        # make sure all masters are listed
        for name in ['M1', 'M2', 'M3']:
            self.assertIn(name, output)
        # make sure keys are not present
        self.assertNotIn('1234', output)
        self.stop_server()

    def test_no_masters(self):
        server = FakeJoule()
        server.response = json.dumps([])
        self.start_server(server)
        runner = CliRunner()
        result = runner.invoke(main, ['master', 'list'])
        print_result_on_error(result)
        self.assertEqual(result.exit_code, 0)
        # should have 3 None values in the output (no Users, Joule, or Lumen nodes)
        self.assertEqual(result.output.count("None"), 3)
        self.stop_server()

    def test_when_server_returns_error_code(self):
        server = FakeJoule()
        error_msg = "test error"
        error_code = 500
        server.response = error_msg
        server.http_code = error_code
        self.start_server(server)
        runner = CliRunner()
        result = runner.invoke(main, ['master', 'list'])
        self.assertIn('%d' % error_code, result.output)
        self.assertIn(error_msg, result.output)
        self.assertEqual(result.exit_code, 1)
        self.stop_server()



import sys
import os
JOULE_PATH = os.path.join(os.path.dirname(__file__), os.pardir, os.pardir, os.pardir)
sys.path.append(JOULE_PATH)
from joule.client import FilterModule
import asyncio


class SimpleFilter(FilterModule):

    async def run(self, parsed_args, inputs, outputs):
        # input1 ----( *2 )---> output1
        input1 = inputs['input1']
        output1 = outputs['output1']
        # input2 ----( *3 )---> output2
        input2 = inputs['input2']
        output2 = outputs['output2']
        while await input1.not_empty():
            data = await input1.read()
            input1.consume(len(data))
            data['data'] *= 2.0
            await output1.write(data)
            if input1.end_of_interval:
                await output1.close_interval()
        while await input2.not_empty():
                data = await input2.read()
                input2.consume(len(data))
                data['data'] *= 3.0
                await output2.write(data)
                if input2.end_of_interval:
                    await output2.close_interval()
        # delay so worker output handler has time to process
        # the results
        await output1.close()
        await output2.close()
        await asyncio.sleep(0.5)


if __name__ == "__main__":
    my_filter = SimpleFilter()
    my_filter.start()



#!/usr/bin/env python3

from joule.utilities import time_now
import joule.client
import asyncio
import numpy as np


class Counter(joule.client.ReaderModule):
    "Counts up from 0 at 10Hz"

    def custom_args(self, parser):
        parser.add_argument("--step", type=int, default=1,
                            help="apply an offset")
        
    async def run(self, parsed_args, output):

        count = 0
        while not self.stop_requested:
            await output.write(np.array([[time_now(), count]]))
            await asyncio.sleep(0.1)
            count += parsed_args.step

            
if __name__ == "__main__":
    r = Counter()
    r.start()



from joule.models.annotation import Annotation
from joule.models.element import Element
from joule.models.event_stream import EventStream
from joule.models.data_stream import DataStream
from joule.models.node import Node
from joule.models.folder import Folder
from joule.models.module import Module
from joule.models.worker import Worker
from joule.models.proxy import Proxy
from joule.models.master import Master
from joule.models.follower import Follower
from joule.models.data_store.data_store import DataStore, StreamInfo, DbInfo
from joule.models.data_store.errors import InsufficientDecimationError, DataError
from joule.models.data_store.event_store import EventStore
from joule.models.data_store.timescale import TimescaleStore
from joule.models.pipes.pipe import Pipe
from joule.models.pipes.local_pipe import LocalPipe
from joule.models.pipes.input_pipe import InputPipe
from joule.models.pipes.output_pipe import OutputPipe
from joule.models.meta import Base



# Apply database migrations based on the difference between the current Joule version and the 
# last version of Joule that accessed this database. This tool is run on daemon startup.

import logging
from sqlalchemy import text
from sqlalchemy.exc import SQLAlchemyError
from joule.version import version as joule_version
from joule.errors import ConfigurationError
import packaging.version
import psycopg2.errors
log = logging.getLogger('joule')

def apply_database_migrations(engine):
    # if this is a new database, no need to run migrations
    if is_new_database(engine):
        log.info("New database, no need to run migrations")
        return
    
    current_version = get_db_version(engine)
    if current_version == packaging.version.parse(joule_version):
        return
    if current_version > packaging.version.parse(joule_version):
        raise ConfigurationError(f"Database version {current_version} is newer than Joule version {joule_version}. This is not supported")
    log.info(f"Detected version: {current_version}, running migrations to version {joule_version} ")
    if current_version < packaging.version.parse("0.11"):
        # Apply migrations for version 0.11
        log.info("Applying migration for version 0.11")
        make_timestamps_timezone_aware(engine)
    else:
        log.info("No migrations to apply")

        
def make_timestamps_timezone_aware(engine):
    with engine.connect() as conn:
        with conn.begin():
            for table,column in [("annotation","start"),
                                ("annotation","end"),
                                ("folder","updated_at"),
                                ("stream","updated_at"),
                                ("eventstream","updated_at"),]:
                conn.execute(text(f"""ALTER TABLE metadata.{table} ALTER COLUMN "{column}" SET DATA TYPE TIMESTAMPTZ  USING "{column}" at time zone 'UTC'"""))

def get_db_version(engine):
    with engine.connect() as conn:
        try:
            row = conn.execute(text("SELECT version FROM metadata.node")).fetchone()
            if row is None:
                return packaging.version.parse("0.0")
            return packaging.version.parse(row[0])
        except SQLAlchemyError:
            log.info("missing table, returning version 0.0")
            return packaging.version.parse("0.0")
       
def is_new_database(engine):
    # if the metadata schema does not exist, no need to run migrations, this is a new database
    with engine.connect() as conn:
        row = conn.execute(text("SELECT 1 FROM pg_namespace WHERE nspname = 'metadata'")).fetchone()
        return row is None
        


from tests.api import mock_session
import unittest
import uuid

from joule.api.node import TcpNode
from joule import errors
import json
from joule.models.data_store.event_store import StreamInfo
from joule.api.event_stream import EventStream, Event
from joule.api.folder import Folder
from joule.errors import ApiError
from joule.constants import EndPoints


class TestEventApi(unittest.IsolatedAsyncioTestCase):

    async def asyncSetUp(self) -> None:
        # no URL or event loop
        self.node = TcpNode('mock_node', 'http://url', 'api_key')
        self.session = mock_session.MockSession()
        self.node.session = self.session

    async def test_event_equality(self):
        self.assertNotEqual(Event(0, 1, content={'a': 'b'}), Event(10, 11, content={'b': 'c'}))
        self.assertEqual(Event(20, 30, content={'a': 'b'}), Event(20, 30, content={'a': 'b'}))

    async def test_event_stream_count(self):
        self.session.response_data = {'count': 10}
        # can count by ID
        await self.node.event_stream_count(1)
        self.assertEqual(self.session.method, 'GET')
        self.assertEqual(self.session.path, EndPoints.event_data_count)
        self.assertEqual(self.session.request_data, {'id': 1, 'include-ongoing-events': 1})

        # can count by path
        await self.node.event_stream_count("/a/path")
        self.assertEqual(self.session.method, 'GET')
        self.assertEqual(self.session.path, EndPoints.event_data_count)
        self.assertEqual(self.session.request_data, {'path': '/a/path', 'include-ongoing-events': 1})

        # can count by EventStream object
        src = EventStream()
        src.id = 1
        await self.node.event_stream_count(src)
        self.assertEqual(self.session.method, 'GET')
        self.assertEqual(self.session.path, EndPoints.event_data_count)
        self.assertEqual(self.session.request_data, {'id': 1, 'include-ongoing-events': 1})

        # can pass optional parameters
        await self.node.event_stream_count(1, start=10, end=20, include_on_going_events=False, json_filter="[[['name','==','test']]]")
        self.assertDictEqual(self.session.request_data, {'start': 10, 'end': 20, 'id': 1,
                                                     'filter': "[[['name','==','test']]]"})

        # handles errors
        with self.assertRaises(errors.ApiError):
            await self.node.event_stream_count(['invalid'])

    async def test_deletes_event_streams(self):
        # can delete by ID
        await self.node.event_stream_delete(1)
        self.assertEqual(self.session.method, 'DELETE')
        self.assertEqual(self.session.path, EndPoints.event)
        self.assertEqual(self.session.request_data, {'id': 1})
        # can delete by path
        await self.node.event_stream_delete('/a/path')
        self.assertEqual(self.session.method, 'DELETE')
        self.assertEqual(self.session.path, EndPoints.event)
        self.assertEqual(self.session.request_data, {'path': '/a/path'})

        # can delete by Folder
        src = EventStream()
        src.id = 1
        await self.node.event_stream_delete(src)
        self.assertEqual(self.session.method, 'DELETE')
        self.assertEqual(self.session.path, EndPoints.event)
        self.assertEqual(self.session.request_data, {'id': 1})

        # handles errors
        with self.assertRaises(errors.ApiError):
            await self.node.event_stream_delete([1, 2, 3])

    async def test_removes_events_from_stream(self):
        # can remove by ID
        await self.node.event_stream_remove(1, start=10, end=20)
        self.assertEqual(self.session.method, 'DELETE')
        self.assertEqual(self.session.path, EndPoints.event_data)
        self.assertEqual(self.session.request_data, {'id': 1, 'start': 10, 'end': 20})

        # can remove by path
        filter_arg = json.dumps([[['name','==','test']]])
        await self.node.event_stream_remove('/a/path', 10, 20, json_filter=filter_arg)
        self.assertEqual(self.session.method, 'DELETE')
        self.assertEqual(self.session.path, EndPoints.event_data)
        self.assertEqual(self.session.request_data, {'path': '/a/path', 'start': 10, 'end': 20, 'filter': filter_arg})

        # can remove by EventStream object
        src = EventStream()
        src.id = 1
        await self.node.event_stream_remove(src, 10, 20)
        self.assertEqual(self.session.method, 'DELETE')
        self.assertEqual(self.session.path, EndPoints.event_data)
        self.assertEqual(self.session.request_data, {'id': 1, 'start': 10, 'end': 20})

        # handles errors
        with self.assertRaises(errors.ApiError):
            await self.node.event_stream_remove([1, 2, 3], 10, 20)

    async def test_creates_event_streams(self):
        src = EventStream(name="test")
        folder = Folder(name="test")
        folder.id = 1
        src_returned = EventStream(name="test")
        src_returned.id = 100
        self.session.response_data = src_returned.to_json()

        # creates with a folder object
        await self.node.event_stream_create(src, folder)
        self.assertEqual(self.session.method, 'POST')
        self.assertEqual(self.session.path, EndPoints.event)
        self.assertEqual(self.session.request_data, {'stream': src.to_json(),
                                                     'dest_id': 1})

        # creates with a folder ID
        await self.node.event_stream_create(src, 10)
        self.assertEqual(self.session.request_data, {'stream': src.to_json(),
                                                     'dest_id': 10})

        # creates with a folder path
        await self.node.event_stream_create(src, "/a/path")
        self.assertEqual(self.session.request_data, {'stream': src.to_json(),
                                                     'dest_path': "/a/path"})

    async def test_gets_event_stream_info(self):
        src = EventStream(name="test")
        src.id = 10
        info = StreamInfo(start=0, end=101, event_count=102, total_time=103, bytes=0)
        self.session.response_data = {'data_info': info.to_json()}

        # gets the info by stream object
        info_returned = await self.node.event_stream_info(src)
        self.assertEqual(self.session.method, 'GET')
        self.assertEqual(self.session.path, EndPoints.event)
        self.assertEqual(self.session.request_data, {'id': 10})
        self.assertEqual(info.start, info_returned.start)
        self.assertEqual(info.end, info_returned.end)
        self.assertEqual(info.event_count, info_returned.event_count)
        self.assertEqual(info.total_time, info_returned.total_time)
        self.assertEqual(info.bytes, info_returned.bytes)

        # gets the info by id
        await self.node.event_stream_info(20)
        self.assertEqual(self.session.request_data, {'id': 20})

        # gets the info by path
        await self.node.event_stream_info("/a/path")
        self.assertEqual(self.session.request_data, {'path': "/a/path"})

        # returns empty info if there is no data for the event stream
        self.session.response_data = {'data_info': None}
        info_returned = await self.node.event_stream_info("/a/path")
        self.assertIsNone(info_returned.start)
        self.assertIsNone(info_returned.end)
        self.assertEqual(info_returned.event_count, 0)
        self.assertEqual(info_returned.total_time, 0)
        self.assertEqual(info_returned.bytes, 0)

    async def test_gets_event_stream(self):
        src = EventStream(name="test")
        src.id = 10
        self.session.response_data = src.to_json()

        # gets the stream by stream object
        src_returned = await self.node.event_stream_get(src)
        self.assertEqual(self.session.method, 'GET')
        self.assertEqual(self.session.path, EndPoints.event)
        self.assertEqual(self.session.request_data, {'id': 10})
        self.assertEqual(src.name, src_returned.name)
        self.assertEqual(src.id, src_returned.id)

        # gets the stream by id
        await self.node.event_stream_get(20)
        self.assertEqual(self.session.request_data, {'id': 20})

        # gets the stream by path
        await self.node.event_stream_get("/a/path")
        self.assertEqual(self.session.request_data, {'path': "/a/path"})

    async def test_creates_event_stream_on_get(self):
        new_stream = EventStream(name="exist")
        new_stream.id = 99
        self.session.response_data = [ApiError("error"), new_stream.to_json()]
        self.session.multiple_calls = True
        src_returned = await self.node.event_stream_get("/does/not/exist", create=True)
        # tries and fails to retrieve the stream
        self.assertEqual(self.session.methods[0], 'GET')
        self.assertEqual(self.session.paths[0], EndPoints.event)
        self.assertEqual(self.session.request_data[0], {'path': "/does/not/exist"})
        # submits another request to create it
        self.assertEqual(self.session.methods[1], 'POST')
        self.assertEqual(self.session.paths[1], EndPoints.event)
        new_stream.id = None
        self.assertEqual(self.session.request_data[1], {'stream': new_stream.to_json(),
                                                        'dest_path': "/does/not"})
        self.assertEqual(src_returned.name, "exist")
        self.assertEqual(src_returned.id, 99)

    async def tests_validates_event_fields(self):
        # valid fields, no exception raised
        EventStream(name="test",
                    event_fields = {'field1': 'string', 'field2': 'numeric', 'field3': 'category:["cat1","cat2"]'})
        

        # invalid fields
        with self.assertRaises(errors.ConfigurationError):
            event_fields = {'field1': 'string', 'field2': 'numeric', 'field3': 'category:[]', 'field4': 'bad'}
            EventStream(name="test",event_fields=event_fields)
        with self.assertRaises(errors.ConfigurationError):
            event_fields = {'field1': 'string', 'field2': 'numeric', 'field3': 'category:bad'}
            EventStream(name="test",event_fields=event_fields)
        with self.assertRaises(errors.ConfigurationError):
            event_fields = {'field1': 'string', 'field2': 'numeric', 'field3': 'bad'}
            EventStream(name="test",event_fields=event_fields)

    async def test_updates_event_stream(self):
        src = EventStream(name="test")
        src.id = 100

        # updates the event stream by object
        await self.node.event_stream_update(src)
        self.assertEqual(self.session.method, 'PUT')
        self.assertEqual(self.session.path, EndPoints.event)
        self.assertEqual(self.session.request_data, {'id': src.id, 'stream': src.to_json()})

        # updating with invalid fields raises an error
        src.event_fields = {'field1': 'bad', 'field2': 'numeric'}
        with self.assertRaises(errors.ConfigurationError):
            await self.node.event_stream_update(src)

    async def test_moves_event_stream(self):
        src = EventStream(name="test")
        src.id = 100
        folder = Folder(name="test")
        folder.id = 1
        await self.node.event_stream_move(src, folder)
        self.assertEqual(self.session.method, 'PUT')
        self.assertEqual(self.session.path, EndPoints.event_move)
        self.assertEqual(self.session.request_data, {'src_id': src.id,
                                                     'dest_id': 1})
        # can move by ID
        await self.node.event_stream_move(10, 100)
        self.assertEqual(self.session.request_data, {'src_id': 10,
                                                     'dest_id': 100})
        # can move by path
        await self.node.event_stream_move("/a/path", "/another/path")
        self.assertEqual(self.session.request_data, {'src_path': "/a/path",
                                                     'dest_path': "/another/path"})

    async def test_writes_events_to_stream(self):
        events = [Event(start_time=i, end_time=i + 1, content={'data': 'value'})
                  for i in range(800)]
        events_returned = [{'id': i, 'event_stream_id': 1, 'start_time': i, 'end_time': i + 1, 'content': {'data': 'value'}}
                           for i in range(800)]
        self.session.response_data = [{'uuid': uuid.uuid4()}, # version.json response
                                      {'events': events_returned[:500]},
                                      {'events': events_returned[500:]}]
        self.session.multiple_calls = True
        await self.node.event_stream_write("/a/path", events)
        self.assertEqual(self.session.methods, ['GET','POST', 'POST'])
        self.assertEqual(self.session.paths, [EndPoints.version_json, EndPoints.event_data, EndPoints.event_data])
        self.assertEqual(self.session.request_data[1]['path'], "/a/path")
        self.assertEqual(len(self.session.request_data[1]['events']), 500)
        self.assertEqual(len(self.session.request_data[2]['events']), len(events[500:]))


import asyncio
import tempfile
import unittest
import datetime
from unittest import mock
import logging
import signal
import stat
import numpy as np
import shlex
import json
from typing import List
import inspect
import os
import argparse
from unittest.mock import Mock
from contextlib import contextmanager
import warnings
from aiohttp import web

from joule.models import Module, DataStream, Worker, Element
from joule.models.supervisor import Supervisor
from joule.models.worker import DataConnection
from joule.models import pipes
from tests import helpers
from joule.constants import EndPoints


LOG_SIZE = 10  # override module default

MODULE_LOG_AND_EXIT = os.path.join(os.path.dirname(__file__),
                                   'worker_scripts', 'log_and_exit.py')
MODULE_IGNORE_SIGTERM = os.path.join(os.path.dirname(__file__),
                                     'worker_scripts', 'ignore_sigterm.py')
MODULE_STOP_ON_SIGTERM = os.path.join(os.path.dirname(__file__),
                                      'worker_scripts', 'stop_on_sigterm.py')
MODULE_ECHO_ARGS = os.path.join(os.path.dirname(__file__),
                                'worker_scripts', 'echo_args.py')
MODULE_SIMPLE_FILTER = os.path.join(os.path.dirname(__file__),
                                    'worker_scripts', 'simple_filter.py')
MODULE_INFINITE_PRINT_LOOP = os.path.join(os.path.dirname(__file__),
                                          'worker_scripts', 'infinite_print_loop.py')
warnings.simplefilter('always')


class TestWorker(unittest.TestCase):

    def setUp(self):
        self.loop = asyncio.new_event_loop()
        self.loop.set_debug(True)
        logging.getLogger('asyncio').setLevel(logging.DEBUG)
        asyncio.set_event_loop(self.loop)
        # generic float32_4 streams
        streams = [DataStream(name="str%d" % n, datatype=DataStream.DATATYPE.FLOAT32,
                              updated_at=datetime.datetime.now(),
                              id=n, elements=[Element(name="e%d" % j, index=j,
                                                      display_type=Element.DISPLAYTYPE.CONTINUOUS) for j in range(3)])
                   for n in
                   range(5)]  # 5th stream is not produced
        self.streams = streams
        # [producer0] --<str0>--,-------------,-<str0,str2>--[consumer0]
        #                       +---[module]--+
        # [producer1] --<str1>--`             `--<str2,str3>--[consumer1]

        self.module = Module(name="module", exec_cmd="/bin/true",
                             description="test module",
                             is_app=False, uuid=123)
        self.module.inputs = {"input1": streams[0], "input2": streams[1]}
        self.module.outputs = {"output1": streams[2], "output2": streams[3]}
        self.module.log_size = LOG_SIZE
        self.worker = Worker(self.module, socket_dir="/tmp", echo_module_logs=False)
        m_producers = [Module(name="producer1", exec_cmd="/bin/runit.sh"),
                       Module(name="producer2", exec_cmd="/bin/runit.sh")]
        m_producers[0].outputs = {"output": streams[0]}
        m_producers[1].outputs = {"output": streams[1]}
        self.producers: List[Worker] = [Worker(m, socket_dir="/tmp", echo_module_logs=False) for m in m_producers]
        m_consumers = [Module(name="consumer1", exec_cmd="/bin/runit.sh"),
                       Module(name="consumer2", exec_cmd="/bin/runit.sh")]
        m_consumers[0].inputs = {"input1": streams[0], "input2": streams[2]}
        m_consumers[1].inputs = {"input1": streams[2], "input2": streams[3]}
        self.consumers: List[Worker] = [Worker(m, socket_dir="/tmp", echo_module_logs=False) for m in m_consumers]
        self.supervisor = Supervisor(self.producers + self.consumers, [], None)

    def tearDown(self):
        closed = self.loop.is_closed()
        if not closed:
            self.loop.call_soon(self.loop.stop)
            self.loop.run_forever()
            self.loop.close()
        asyncio.set_event_loop(None)

    def test_builds_worker_from_module(self):
        # subscriber arrays are empty
        self.assertEqual(self.worker.subscribers, {self.streams[2]: [], self.streams[3]: []})
        # data connections are empty
        self.assertEqual(self.worker.input_connections, [])
        # output connections are empty
        self.assertEqual(self.worker.output_connections, [])

    @mock.patch('joule.models.worker.get_stream_path')
    def test_provides_module_attributes(self, mock_path: mock.Mock):
        mock_path.return_value = "/mock/path"
        self.assertEqual(self.worker.uuid, self.module.uuid)
        self.assertEqual(self.worker.name, self.module.name)
        self.assertEqual(self.worker.description, self.module.description)
        self.assertEqual(self.worker.is_app, self.module.is_app)
        connection = DataConnection("stub", 0, self.streams[0], pipes.Pipe())
        self.assertEqual(connection.location, "/mock/path")

    def test_produces_returns_true_if_worker_makes_output(self):

        # an output
        self.assertTrue(self.worker.produces(self.streams[2]))
        # an input
        self.assertFalse(self.worker.produces(self.streams[0]))
        # an unrelated stream
        s = helpers.create_stream("unrelated", "int16_10")
        self.assertFalse(self.worker.produces(s))

    def test_spawns_child_process(self):

        asyncio.run(self.worker.run(self.supervisor.subscribe, restart=False))
        self.assertEqual(self.worker.process.returncode, 0)

    def test_restarts_child(self):
        self.worker.RESTART_INTERVAL = 0.2

        # subscribe to the module outputs
        output1 = pipes.LocalPipe(layout=self.streams[2].layout)
        self.worker.subscribe(self.streams[2], output1)

        async def runner():
            task = asyncio.create_task(self.worker.run(self.supervisor.subscribe, restart=True))
            await asyncio.sleep(0.5)
            await self.worker.stop()
            await task

        with self.assertLogs(level="WARNING"):
            with self.check_fd_leakage():
                asyncio.run(runner())

        self.assertEqual(self.worker.process.returncode, 0)
        self.assertEqual(len(output1.read_nowait()), 0)
        self.assertTrue(output1.end_of_interval)

    def test_collects_statistics(self):
        # child should listen for stop_requested flag

        self.module.exec_cmd = "python3 " + MODULE_STOP_ON_SIGTERM

        async def get_statistics():
            await asyncio.sleep(0.1)
            statistics = await self.worker.statistics()
            self.assertIsNotNone(statistics.pid)
            self.assertGreater(statistics.memory_percent, 0)
            # kill the process and try to get statistics again
            os.kill(statistics.pid, signal.SIGKILL)
            await asyncio.sleep(0.1)
            null_statistics = await self.worker.statistics()
            self.assertIsNone(null_statistics.pid)
            self.assertIsNone(null_statistics.memory_percent)

        # no statistics available before worker starts
        stats = asyncio.run(self.worker.statistics())
        self.assertIsNone(stats.pid)

        async def runner():
            await asyncio.gather(
                get_statistics(),
                self.worker.run(self.supervisor.subscribe, restart=False))

        with self.check_fd_leakage():
            asyncio.run(runner())

    def test_restarts_and_stops_child_by_request(self):
        # child should listen for stop_requested flag
        # loop = asyncio.get_event_loop()
        self.worker.RESTART_INTERVAL = 0.01
        self.module.exec_cmd = "python3 " + MODULE_STOP_ON_SIGTERM

        async def runner():
            # calling stop before run doesn't matter
            await self.worker.stop()
            task = asyncio.create_task(self.worker.run(self.supervisor.subscribe, restart=True))
            await asyncio.sleep(0.2)
            await self.worker.restart()
            await asyncio.sleep(0.2)
            await self.worker.stop()
            await task

        with self.assertLogs(level="WARNING") as logs:
            with self.check_fd_leakage():
                asyncio.run(runner())
                # asyncio.run(asyncio.gather(
                #    loop.create_task(self._stop_worker(loop)),
                #    loop.create_task(self._restart_worker(loop)),
                #    loop.create_task()
                # ))
        # check to make sure it was killed by SIGTERM
        self.assertEqual(self.worker.process.returncode, -1 * signal.SIGTERM)
        # the restart should be logged
        self.assertIn("restarting", ''.join(logs.output).lower())
        # make the the module started multiple times
        num_starts = 0
        for entry in self.worker.logs:
            if 'starting' in entry:
                num_starts += 1
        self.assertEqual(num_starts, 2)
        # can call stop multiple times
        asyncio.run(self.worker.stop())

    def test_terminates_child(self):
        # send SIGKILL to terminate bad children
        self.module.exec_cmd = "python3 " + MODULE_IGNORE_SIGTERM
        # speed up the test
        self.worker.SIGTERM_TIMEOUT = 0.5

        async def runner():
            task = asyncio.create_task(self.worker.run(self.supervisor.subscribe, restart=False))
            await asyncio.sleep(0.5)
            await self.worker.stop()
            await asyncio.sleep(2)

        with self.assertLogs(level="WARNING"):
            with self.check_fd_leakage():
                asyncio.run(runner())
        # check to make sure it was killed by SIGKILL
        self.assertEqual(self.worker.process.returncode, -1 * signal.SIGKILL)

    def test_builds_child_arguments(self):

        self.module.exec_cmd = "python3 " + MODULE_ECHO_ARGS
        self.module.arguments = {'arg1': "value1",
                                 'arg2': "value2"}
        self.module.is_app = True
        self.worker.log = Mock()
        asyncio.run(self.worker.run(self.supervisor.subscribe, restart=False))
        self.assertEqual(self.worker.process.returncode, 0)
        # expect to get a pipes argument that is a json string
        parser = argparse.ArgumentParser()
        parser.add_argument("--pipes")
        parser.add_argument("--socket")
        parser.add_argument("--api-socket")
        parser.add_argument("--node")
        parser.add_argument("--arg1")
        parser.add_argument("--arg2")
        # get the second log entry which is the echo'd arguments
        argv = shlex.split(self.worker.log.mock_calls[1][1][0])
        args = parser.parse_args(argv)
        my_pipes = json.loads(args.pipes)
        socket_name = args.socket
        # verify inputs and outputs are in the config
        value = self.streams[0].id
        self.assertEqual(my_pipes['inputs']['input1']['id'], value)
        value = self.streams[1].id
        self.assertEqual(my_pipes['inputs']['input2']['id'], value)
        value = self.streams[2].id
        self.assertEqual(my_pipes['outputs']['output1']['id'], value)
        value = self.streams[3].id
        self.assertEqual(my_pipes['outputs']['output2']['id'], value)
        self.assertEqual(socket_name, self.worker.app_socket)
        self.assertEqual(args.arg1, "value1")
        self.assertEqual(args.arg2, "value2")

    def test_logs_child_output(self):

        self.module.description = "test"
        self.module.exec_cmd = "python3 " + MODULE_LOG_AND_EXIT
        asyncio.run(self.worker.run(self.supervisor.subscribe, restart=False))
        self.assertEqual(self.worker.process.returncode, 0)
        # log should have a starting entry
        self.assertRegex(self.worker.logs[0], 'starting')
        # ...content produced by the module (just echo'd params)
        self.assertRegex(self.worker.logs[1], 'hello world')
        # ...and a terminating entry
        self.assertRegex(self.worker.logs[2], 'terminated')

    def test_inputs_must_be_available(self):
        self.module.inputs["missing_input"] = self.streams[4]

        with self.assertLogs(level="ERROR"):
            asyncio.run(self.worker.run(self.supervisor.subscribe, restart=False))
        for entry in self.worker.logs:
            if "inputs are not available" in entry:
                break
        else:
            self.fail("missing log entry")

    def test_handles_invalid_exec_cmds(self):
        self.module.exec_cmd = "bad-cmd"

        with self.assertLogs(level="ERROR"):
            with self.check_fd_leakage():
                asyncio.run(self.worker.run(self.supervisor.subscribe, restart=False))
        for entry in self.worker.logs:
            if "cannot start module" in entry:
                break
        else:
            self.fail("missing log entry")

    def test_rolls_logs(self):
        self.module.exec_cmd = "/usr/bin/env python3 " + MODULE_INFINITE_PRINT_LOOP

        async def runner():
            task = asyncio.create_task(self.worker.run(self.supervisor.subscribe, restart=False))
            await asyncio.sleep(0.5)
            await self.worker.stop()

        with self.check_fd_leakage():
            asyncio.run(runner())
        logs = self.worker.logs
        self.assertEqual(len(logs), LOG_SIZE)
        # make sure the entries have rolled
        self.assertNotIn('starting', logs[0])
        # make sure the last entry is the most recent
        self.assertIn('terminated', logs[-1])

    """
            Filter accepts two inputs and produces two outputs
            Streams 0,1 are inputs produced by mock_producers
            Streams 2,3 are produced by the filter
            Output1 is sent to
            """

    def test_passes_data_across_pipes(self):

        # create worker connections
        # child runs until stopped
        self.module.exec_cmd = "/usr/bin/env python3 " + MODULE_SIMPLE_FILTER

        interval1_data = helpers.create_data('float32_3', start=1000, step=100, length=100)
        interval2_data = helpers.create_data('float32_3', start=1001 + 100 * 100, step=100, length=100)

        # create a stub server to respond to API calls as the module starts up
        app = web.Application()

        node_stream_info_api_call_count = 0

        async def stub_stream_info(request):
            nonlocal node_stream_info_api_call_count
            node_stream_info_api_call_count += 1
            stream_id = int(request.query['id'])
            return web.json_response(self.streams[stream_id].to_json())

        app.add_routes([web.get(EndPoints.stream, stub_stream_info)])
        web_runner = web.AppRunner(app)
        tmp_dir = tempfile.TemporaryDirectory()
        sock_file = os.path.join(tmp_dir.name, 'testing')
        self.worker.api_socket = sock_file

        async def mock_producers():
            # await asyncio.sleep(0.5)
            # wait for filter to subscribe to input stream0
            while True:
                if len(self.producers[0].subscribers[self.streams[0]]) > 0:
                    break
                await asyncio.sleep(0.01)
            stream = self.producers[0].subscribers[self.streams[0]][0]
            await stream.write(interval1_data)
            await stream.close_interval()
            await stream.write(interval2_data)
            await stream.close_interval()
            await stream.close()

            # wait for filter to subscribe to input 1
            while True:
                if len(self.producers[1].subscribers[self.streams[1]]) > 0:
                    break
                await asyncio.sleep(0.01)
            stream = self.producers[1].subscribers[self.streams[1]][0]
            await stream.write(interval1_data)
            await stream.close_interval()
            await stream.write(interval2_data)
            await stream.close_interval()
            await stream.close()

            await asyncio.sleep(2)

        # subscribe to the module outputs
        output1 = pipes.LocalPipe(layout=self.streams[2].layout, name="output1", debug=False)
        output2 = pipes.LocalPipe(layout=self.streams[3].layout, name="output2", debug=False)

        # create a slow subscriber that times out
        class SlowPipe(pipes.Pipe):
            async def write(self, data):
                await asyncio.sleep(10)

            async def close_interval(self):
                # API compatability
                pass

        slow_pipe = SlowPipe(stream=helpers.create_stream('slow stream', self.streams[2].layout), name='slow pipe')
        self.worker.subscribe(self.streams[2], slow_pipe)
        self.worker.SUBSCRIBER_TIMEOUT = 0.1

        # create a subscriber that errors out
        class ErrorPipe(pipes.Pipe):
            async def write(self, data):
                raise BrokenPipeError()

        error_pipe = ErrorPipe(stream=helpers.create_stream('error stream', self.streams[2].layout), name='error pipe')
        self.worker.subscribe(self.streams[3], error_pipe)
        self.worker.subscribe(self.streams[3], output2)

        self.worker.subscribe(self.streams[2], output1)

        async def runner():
            await web_runner.setup()
            sock_site = web.UnixSite(web_runner, sock_file)
            await sock_site.start()

            await asyncio.gather(mock_producers(),
                                 self.worker.run(self.supervisor.subscribe, restart=False))

            await web_runner.shutdown()
            await web_runner.cleanup()

        with self.assertLogs() as log:
            asyncio.run(runner())

        # make sure the module queried the API endpoint for each stream
        self.assertEqual(node_stream_info_api_call_count, 4)
        tmp_dir.cleanup()  # remove socket file and directory
        log_dump = '\n'.join(log.output)
        self.assertIn("subscriber disconnected", log_dump)
        self.assertIn("timed out", log_dump)
        # check stream2, should be stream0*2.0 [] stream0*2.0
        output_data = output1.read_nowait()
        output1.consume(len(output_data))
        np.testing.assert_array_almost_equal(interval1_data['data'] * 2.0,
                                             output_data['data'])
        self.assertTrue(output1.end_of_interval)
        output_data = output1.read_nowait()
        output1.consume(len(output_data))
        np.testing.assert_array_almost_equal(interval2_data['data'] * 2.0,
                                             output_data['data'])
        self.assertTrue(output1.end_of_interval)

        # check stream3, should be stream1*3.0 [] stream1*3.0
        output_data = output2.read_nowait()
        output2.consume(len(output_data))
        np.testing.assert_array_almost_equal(interval1_data['data'] * 3.0,
                                             output_data['data'])
        self.assertTrue(output2.end_of_interval)
        output_data = output2.read_nowait()
        output2.consume(len(output_data))
        np.testing.assert_array_almost_equal(interval2_data['data'] * 3.0,
                                             output_data['data'])
        self.assertTrue(output2.end_of_interval)

    @contextmanager
    def check_fd_leakage(self):
        existing_pipes = 0
        # find out how many pipes are currently open
        for str_fd in set(os.listdir('/proc/self/fd/')):
            fd = int(str_fd)
            try:
                fstat = os.fstat(fd)
                if stat.S_ISFIFO(fstat.st_mode):
                    existing_pipes += 1
            except OSError:
                pass

        yield
        # make sure the same number of pipes are open
        new_pipes = 0
        for str_fd in set(os.listdir('/proc/self/fd/')):
            fd = int(str_fd)
            try:
                fstat = os.fstat(fd)
                if stat.S_ISFIFO(fstat.st_mode):
                    new_pipes += 1
            except OSError:
                pass

        self.assertEqual(existing_pipes, new_pipes)


""" from git gist """
descriptors = set()


def print_open_fds(print_all=False):
    global descriptors
    (frame, filename, line_number, function_name, lines, index) = inspect.getouterframes(inspect.currentframe())[1]
    fds = set(os.listdir('/proc/self/fd/'))
    new_fds = fds - descriptors
    closed_fds = descriptors - fds
    descriptors = fds

    if print_all:
        print("{}:{} ALL file descriptors: {}".format(filename, line_number, fds))

    if new_fds:
        print("{}:{} new file descriptors: {}".format(filename, line_number, new_fds))
    if closed_fds:
        print("{}:{} closed file descriptors: {}".format(filename, line_number, closed_fds))



import unittest

from joule.utilities import time


class TestTimeUtilities(unittest.TestCase):

    def test_time_helpers(self):
        now = time.time_now()
        time_str = time.timestamp_to_human(now)
        # make sure it doesn't raise any errors
        self.assertGreater(len(time_str), 0)
        # check special timestamps
        self.assertEqual("(minimum)", time.timestamp_to_human(time.min_timestamp))
        self.assertEqual("(maximum)", time.timestamp_to_human(time.max_timestamp))
        # check conversion functions
        unix_now = time.timestamp_to_unix(now)
        self.assertEqual(now, time.unix_to_timestamp(unix_now))



from tests.helpers import DbTestCase
from joule.errors import ConfigurationError
from joule.services import parse_pipe_config


class TestParsePipeConfigErrors(DbTestCase):

    def test_ensures_valid_path_and_name(self):
        bad_configs = [
            "/no_path:int16[x,y]",
            "missing/slash:int16[x,y]",
            "justbad",
            "",
            "/tooshort"
            "/no/data/format",
            "/empty/format:[]",
            "/inval:d/characters",
            "/bad/elements:bad[x,y]",
            "/bad/elements:float32[x,",
            "/bad/elements:float32[x,y",
            "/bad/elements:float32[,",
            "//too/many//slashes",
            "remote.stream/without/space"
            "remote:3000 /no/inline_config"
        ]
        for config in bad_configs:
            with self.assertRaises(ConfigurationError):
                parse_pipe_config.run(config, self.db)



from joule.models.pipes import LocalPipe
import unittest
import asyncio
import numpy as np
import argparse

from tests import helpers
from joule.client.builtins.mean_filter import MeanFilter

WINDOW = 9
WIDTH = 4
REPS_PER_BLOCK = 9
NUM_BLOCKS = 3


class TestMeanFilter(helpers.AsyncTestCase):

    def test_computes_moving_average(self):
        # data is a repeating series of 0,1,2,..N
        # the moving average of this array should be N/2
        # timestamps is just an increasing index

        my_filter = MeanFilter()
        pipe_in = LocalPipe("float32_%d" % WIDTH, name="input")
        pipe_out = LocalPipe("float32_%d" % WIDTH, name="output")
        args = argparse.Namespace(window=WINDOW, pipes="unset")
        base = np.array([np.arange(x, WINDOW + x) for x in range(WIDTH)]).T

        async def writer():
            prev_ts = 0
            for block in range(NUM_BLOCKS):
                data = np.tile(base, (REPS_PER_BLOCK, 1))
                ts = np.arange(prev_ts, prev_ts + len(data))
                input_block = np.hstack((ts[:, None], data))
                pipe_in.write_nowait(input_block)
                # await asyncio.sleep(0.1)
                prev_ts = ts[-1] + 1
            # now put in an extra block after an interval break
            await pipe_in.close_interval()
            # all 1's (even timestamps)
            await pipe_in.write(np.ones((100, WIDTH + 1)))
            # await asyncio.sleep(0.2)
            await pipe_in.close()

        # run reader in an event loop
        async def runner():
            await asyncio.gather(writer(),
                           my_filter.run(args,
                                         {"input": pipe_in},
                                         {"output": pipe_out}))
        asyncio.run(runner())
        # check the results
        result = pipe_out.read_nowait()
        # expect the output to be a constant (WINDOW-1)/2
        base_output = np.ones((WINDOW * REPS_PER_BLOCK *
                               NUM_BLOCKS - (WINDOW - 1), WIDTH))
        expected = base_output * range(int(WINDOW / 2), int(WINDOW / 2) + WIDTH)
        np.testing.assert_array_equal(expected,
                                      result['data'])
        self.assertTrue(pipe_out.end_of_interval)
        pipe_out.consume(len(result))
        # now read the extra block and make sure it has all the data
        result = pipe_out.read_nowait(flatten=True)
        self.assertEqual(len(result), 100 - WINDOW + 1)



from tests.api import mock_session
import unittest

from joule.api.node import TcpNode
from joule import errors

from joule.api.data_stream import DataStream
from joule.api.annotation import Annotation


class TestAnnotationApi(unittest.IsolatedAsyncioTestCase):

    async def asyncSetUp(self):
        # no URL or event loop
        self.node = TcpNode('mock_node', 'http://url', 'api_key')
        self.session = mock_session.MockSession()
        self.node.session = self.session

    async def test_annotation_create(self):
        annotation = Annotation(title="note",
                                start=100,
                                content="nothing special")
        json = {
            "id": 1,
            "stream_id": 1,
            "title": "nozte",
            "start": 100,
            "end": None,
            "content": "nothing special"}
        self.session.response_data = json
        # create by stream path
        await self.node.annotation_create(annotation, '/a/path')
        self.assertEqual(self.session.request_data["stream_path"], "/a/path")
        stream = DataStream(name="test")
        stream.id = 100
        # create by DataStream object
        await self.node.annotation_create(annotation, stream)
        self.assertEqual(self.session.request_data["stream_id"], 100)
        # create by DataStream id
        await self.node.annotation_create(annotation, 100)
        self.assertEqual(self.session.request_data["stream_id"], 100)
        # errors with invalid stream type
        with self.assertRaises(errors.ApiError):
            await self.node.annotation_create(annotation, [1, 2])

    async def test_annoation_delete(self):
        annotation = Annotation(title="note",
                                start=100,
                                content="nothing special")
        annotation.id = 96
        self.session.response_data = ""
        # can delete by Annotation object
        await self.node.annotation_delete(annotation)
        self.assertEqual(self.session.request_data["id"], 96)
        # can delete by Annotation id
        await self.node.annotation_delete(annotation.id)
        self.assertEqual(self.session.request_data["id"], 96)
        # errors with invalid annotation type
        with self.assertRaises(errors.ApiError):
            await self.node.annotation_create(annotation, [1, 2])

    async def test_annotation_update(self):
        annotation = Annotation(title="note",
                                start=100,
                                content="nothing special")
        annotation.id = 1
        json = {
            "id": 1,
            "stream_id": 1,
            "title": "note",
            "start": 100,
            "end": None,
            "content": "nothing special"}
        self.session.response_data = json
        await self.node.annotation_update(annotation)
        self.assertEqual(self.session.request_data["id"], 1)

    async def test_annotation_get(self):
        json = [{
            "id": 1,
            "stream_id": 1,
            "title": "note",
            "start": 100,
            "end": None,
            "content": "nothing special"}]

        # get by stream id
        self.session.response_data = json
        await self.node.annotation_get(1)
        req_data = self.session.request_data
        self.assertEqual(req_data["stream_id"], 1)

        # get by stream object
        stream = DataStream(name="test")
        stream.id = 100
        await self.node.annotation_get(stream)
        req_data = self.session.request_data
        self.assertEqual(req_data["stream_id"], 100)

        # get by stream path
        await self.node.annotation_get("/stream/path")
        req_data = self.session.request_data
        self.assertEqual(req_data["stream_path"], "/stream/path")

        # error otherwise
        with self.assertRaises(errors.ApiError):
            await self.node.annotation_get({})












from joule.models.pipes import LocalPipe
import asyncio
import numpy as np
import argparse

from tests import helpers
from joule.client.builtins.random_reader import RandomReader

WIDTH = 2
RATE = 100


class TestRandomReader(helpers.AsyncTestCase):

    def test_generates_random_values(self):

        my_reader = RandomReader()
        loop = asyncio.new_event_loop()
        pipe = LocalPipe("float32_%d" % WIDTH, name="output")
        args = argparse.Namespace(width=WIDTH, rate=RATE, pipes="unset")

        loop.call_later(0.1, my_reader.stop)
        loop.run_until_complete(my_reader.run(args, pipe))
        loop.close()
        # check the results
        result = pipe.read_nowait()
        diffs = np.diff(result['timestamp'])
        self.assertEqual(np.mean(diffs), 1 / RATE * 1e6)
        self.assertEqual(np.shape(result['data'])[1], WIDTH)



from aiohttp.test_utils import AioHTTPTestCase, unittest_run_loop
from aiohttp import web
from sqlalchemy.orm import Session
import asyncio
from unittest import mock
from joule.models import master, Master
import joule.controllers
from tests.controllers.helpers import create_db, MockStore
from joule import app_keys
from joule.constants import EndPoints

import testing.postgresql
psql_key = web.AppKey("psql", testing.postgresql.Postgresql)

class TestMasterController(AioHTTPTestCase):

    async def tearDownAsync(self):
        self.app[app_keys.db].close()
        await asyncio.to_thread(self.app[psql_key].stop)
        await self.client.close()

    async def get_application(self):
        app = web.Application()
        app.add_routes(joule.controllers.routes)
        # this takes a while, adjust the expected coroutine execution time
        loop = asyncio.get_running_loop()
        loop.slow_callback_duration = 2.0

        app[app_keys.db], app[psql_key] = await asyncio.to_thread(lambda: create_db([]))
        app[app_keys.data_store] = MockStore()
        app[app_keys.name] = "test"
        app[app_keys.port] = 443
        app[app_keys.scheme] = "http"
        app[app_keys.base_uri] = "/"
        app[app_keys.cafile] = "stubbed_cafile"

        # create a master user to grant access
        self.grantor = Master(name="grantor", key="grantor_key",
                              type=Master.TYPE.USER)
        app[app_keys.db].add(self.grantor)
        app[app_keys.db].commit()
        return app


    async def test_adds_master_user(self):
        db: Session = self.app[app_keys.db]
        # add user to johndoe
        payload = {
            "master_type": "user",
            "identifier": "johndoe"
        }
        resp = await self.client.post(EndPoints.master, json=payload,
                                      headers={'X-API-KEY': "grantor_key"})
        self.assertEqual(resp.status, 200)
        json = await resp.json()
        m = db.query(Master).filter(Master.name == "johndoe").one_or_none()
        self.assertEqual(m.grantor_id, self.grantor.id)
        self.assertEqual(json['key'], m.key)

        # adding johndoe again returns an error and does not change the key
        resp = await self.client.post(EndPoints.master, json=payload,
                                      headers={'X-API-KEY': "grantor_key"})
        old_key = m.key
        m = db.query(Master).filter(Master.name == "johndoe").one_or_none()
        new_key = m.key
        self.assertEqual(old_key, new_key)
        self.assertEqual(resp.status, 400)


    @mock.patch('joule.controllers.master_controller.TcpSession')
    @mock.patch('joule.controllers.master_controller.detect_url')
    async def test_adds_master_joule(self, detect_url, mock_session_class):
        mock_session = mock.AsyncMock()
        mock_session_class.return_value = mock_session
        # mock the post coroutine to return the expected JSON data
        mock_session.post.return_value = {"name": "joule_master_node"}

        db: Session = self.app[app_keys.db]
        # add joule node
        payload = {
            "master_type": "joule",
            "identifier": "http://master_node",
        }
        resp = await self.client.post(EndPoints.master, json=payload,
                                        headers={'X-API-KEY': "grantor_key"})
        # check the response
        self.assertEqual(resp.status, 200)
        data = await resp.json()
        # make sure the database entry is created
        self.assertEqual(db.query(Master).filter(Master.name == "joule_master_node").count(),1)
        self.assertEqual(data['name'], "joule_master_node")
        # make sure the other node received the correct data
        mock_session.post.assert_called_once()
        self.assertEqual(mock_session.post.call_args[0][0], EndPoints.follower)
        params = mock_session.post.call_args[1]['json']
        self.assertIn('key', params)
        self.assertEqual(params['port'], self.app[app_keys.port])
        self.assertEqual(params['scheme'], self.app[app_keys.scheme])
        self.assertEqual(params['base_uri'], self.app[app_keys.base_uri])
        self.assertEqual(params['name'], "test")

        # if the master is not specified with a URL, try to create one
        detect_url.return_value = "https://master_node"
        payload = {
            "master_type": "joule",
            "identifier": "master_node",
        }
        resp = await self.client.post(EndPoints.master, json=payload,
                                        headers={'X-API-KEY': "grantor_key"})
        self.assertEqual(resp.status, 200)
        mock_session_class.call_count = 2
        self.assertEqual("https://master_node",mock_session_class.call_args[0][0])

    @mock.patch('joule.controllers.master_controller.TcpSession')
    @mock.patch('joule.controllers.master_controller.detect_url')
    async def test_adds_master_lumen(self, detect_url, mock_session_class):
        mock_session = mock.AsyncMock()
        mock_session_class.return_value = mock_session
        # mock the post coroutine to return the expected JSON data
        mock_session.post.return_value = {"name": "lumen_master_node"}

        db: Session = self.app[app_keys.db]
        # add joule node
        payload = {
            "master_type": "lumen",
            "identifier": "http://master_node",
            "lumen_params":{"key":"stubbed"}
        }
        resp = await self.client.post(EndPoints.master, json=payload,
                                        headers={'X-API-KEY': "grantor_key"})
        # check the response
        self.assertEqual(resp.status, 200)
        data = await resp.json()
        # make sure the database entry is created
        self.assertEqual(db.query(Master).filter(Master.name == "http://master_node").count(),1)
        self.assertEqual(data['name'], "http://master_node")
        # make sure the lumen node received the correct data
        mock_session.post.assert_called_once()
        self.assertEqual(mock_session.post.call_args[0][0], "/nilms.json")
        params = mock_session.post.call_args[1]['json']
        self.assertIn('key', params)
        self.assertEqual(params['port'], self.app[app_keys.port])
        self.assertEqual(params['scheme'], self.app[app_keys.scheme])
        self.assertEqual(params['base_uri'], self.app[app_keys.base_uri])
        self.assertEqual(params['name'], "test")

        # if the master is not specified with a URL, try to create one
        detect_url.return_value = "https://master_node"
        payload = {
            "master_type": "lumen",
            "identifier": "master_node",
            "lumen_params":{"key":"stubbed"}
        }
        resp = await self.client.post(EndPoints.master, json=payload,
                                        headers={'X-API-KEY': "grantor_key"})
        self.assertEqual(resp.status, 200)
        mock_session_class.call_count = 2
        self.assertEqual("https://master_node",mock_session_class.call_args[0][0])


    async def test_removes_master_user(self):
        db: Session = self.app[app_keys.db]
        db.add(Master(name="johndoe", key=master.make_key(), type=Master.TYPE.USER))
        db.commit()
        m = db.query(Master).filter(Master.name == "johndoe").one_or_none()
        self.assertIsNotNone(m)

        # can remove users
        payload = {
            "master_type": "user",
            "name": "johndoe"
        }
        resp = await self.client.delete(EndPoints.master, params=payload,
                                        headers={'X-API-KEY': "grantor_key"})
        self.assertEqual(resp.status, 200)
        m = db.query(Master).filter(Master.name == "johndoe").one_or_none()
        self.assertIsNone(m)

        # can remove nodes
        db.add(Master(name="joule", key=master.make_key(), type=Master.TYPE.JOULE_NODE))
        db.commit()
        m = db.query(Master).filter(Master.name == "joule").one_or_none()
        self.assertIsNotNone(m)
        payload = {
            "master_type": "joule",
            "name": "joule"
        }
        resp = await self.client.delete(EndPoints.master, params=payload,
                                        headers={'X-API-KEY': "grantor_key"})
        self.assertEqual(resp.status, 200)
        m = db.query(Master).filter(Master.name == "joule").one_or_none()
        self.assertIsNone(m)


    async def test_lists_masters(self):
        db: Session = self.app[app_keys.db]
        db.add(Master(name="johndoe", key=master.make_key(), type=Master.TYPE.USER))
        db.add(Master(name="joule", key=master.make_key(), type=Master.TYPE.JOULE_NODE))
        db.add(Master(name="lumen", key=master.make_key(), type=Master.TYPE.LUMEN_NODE))
        db.commit()

        resp = await self.client.get(EndPoints.masters,
                                     headers={'X-API-KEY': "grantor_key"})
        self.assertEqual(resp.status, 200)
        json = await resp.json()
        # grantor is the original user
        self.assertEqual(len(json), 4)
        for item in json:
            self.assertNotIn("key", item)
            self.assertIn(item["name"], ["johndoe", "joule", "lumen", "grantor"])



import unittest
from joule.models import (Element, DataStream, Folder)


class TestSchema(unittest.TestCase):
    def test_relationships(self):
        # root
        #  -folder1
        #     -stream11 (4 elements)
        #  -folder2
        #     -stream21 (2 elements)
        #  -stream1 (1 element)
        #  -stream2 (1 element)

        stream11 = DataStream(name="stream11", datatype=DataStream.DATATYPE.FLOAT32)
        stream11.elements = [Element(name="e%d" % x,
                                     display_type=Element.DISPLAYTYPE.DISCRETE) for x in range(4)]
        folder1 = Folder(name="folder1")
        folder1.data_streams.append(stream11)

        stream21 = DataStream(name="stream21", datatype=DataStream.DATATYPE.INT16)
        stream21.elements = [Element(name="e%d" % x,
                                     display_type=Element.DISPLAYTYPE.CONTINUOUS) for x in range(4)]
        folder2 = Folder(name="folder2")
        folder2.data_streams.append(stream21)

        stream1 = DataStream(name="stream1", datatype=DataStream.DATATYPE.INT16)
        stream1.elements.append(Element(name="e0"))

        stream2 = DataStream(name="stream2", datatype=DataStream.DATATYPE.INT64)
        stream2.elements.append(Element(name="e0"))

        root = Folder(name="root")
        root.children = [folder1, folder2]
        root.data_streams = [stream1, stream2]
        # check downward navigation
        self.assertEqual(len(root.children[0].data_streams[0].elements), 4)
        # check upward navigation
        e = stream11.elements[-1]
        self.assertEqual(e.stream.folder.parent.name, 'root')



from typing import Union, List, Optional, Dict, TYPE_CHECKING, Type
from joule.constants import EndPoints

if TYPE_CHECKING:
    import sqlalchemy
    from joule.api.data_stream import DataStream
    from joule.api.event_stream import EventStream, EventStreamInfo
    from joule.api.module import Module
    from joule.api.proxy import Proxy
    from joule.api.master import Master
    from joule.api.annotation import Annotation, AnnotationInfo
    from joule.api.folder import Folder
    from joule.api.event import Event
    from joule.api.data_stream import DataStreamInfo, DataStream
    from joule.models.pipes import Pipe
    from joule.utilities import ConnectionInfo
    from joule.api.session import BaseSession
    from .node_info import NodeInfo


class BaseNode:

    def __init__(self, name: str,
                 session: 'BaseSession'):
        self.name = name
        self.session = session
        # configured in child implementations
        self.url = "<unset>"

    async def close(self):
        await self.session.close()

    async def info(self) -> 'NodeInfo':
        from .node_info import NodeInfo
        resp = await self.session.get(EndPoints.version_json)
        db_info = await self.session.get(EndPoints.db_info)
        return NodeInfo(version=resp["version"], name=resp["name"], path=db_info['path'],
                        size_other=db_info["other"], size_reserved=db_info["reserved"],
                        size_free=db_info["free"], size_db=db_info["size"])

    # Folder actions

    async def folder_root(self) -> 'Folder':
        from joule.api.folder import folder_root
        return await folder_root(self.session)

    async def folder_get(self,
                         folder: 'Folder | str | int') -> 'Folder':
        from joule.api.folder import folder_get
        return await folder_get(self.session, folder)

    async def folder_move(self,
                          source: 'Folder | str | int',
                          destination: 'Folder | str | int') -> None:
        from joule.api.folder import folder_move
        await folder_move(self.session, source, destination)

    async def folder_update(self,
                            folder: 'Folder') -> None:
        from joule.api.folder import folder_update
        await folder_update(self.session, folder)

    async def folder_delete(self,
                            folder: 'Folder | str | int',
                            recursive: bool = False) -> None:
        from joule.api.folder import folder_delete
        await folder_delete(self.session, folder, recursive)

    # DataStream actions

    async def data_stream_get(self,
                              stream: 'DataStream | str | int') -> 'DataStream':
        from joule.api.data_stream import data_stream_get
        return await data_stream_get(self.session, stream)

    async def data_stream_move(self,
                               stream: 'DataStream | str | int',
                               folder: 'Folder | str | int') -> None:
        from joule.api.data_stream import data_stream_move
        return await data_stream_move(self.session, stream, folder)

    async def data_stream_update(self,
                                 stream: 'DataStream') -> None:
        from joule.api.data_stream import data_stream_update
        return await data_stream_update(self.session,
                                        stream)

    async def data_stream_delete(self,
                                 stream: 'DataStream | str | int') -> None:
        from joule.api.data_stream import data_stream_delete
        await data_stream_delete(self.session, stream)

    async def data_stream_create(self,
                                 stream: 'DataStream',
                                 folder: 'Folder | str | int') -> 'DataStream':
        from joule.api.data_stream import data_stream_create
        return await data_stream_create(self.session, stream, folder)

    async def data_stream_info(self,
                               stream: 'DataStream | str | int') -> 'DataStreamInfo':
        from joule.api.data_stream import data_stream_info
        return await data_stream_info(self.session, stream)

    async def data_stream_annotation_delete(self,
                                            stream: 'DataStream | str | int',
                                            start: Optional[int] = None,
                                            end: Optional[int] = None):
        from joule.api.data_stream import data_stream_annotation_delete
        return await data_stream_annotation_delete(self.session, stream, start, end)

    # EventStream actions

    async def event_stream_get(self,
                               stream: 'EventStream | str | int',
                               create: bool = False,
                               description: str = "",
                               chunk_duration: str = "",
                               chunk_duration_us: Optional[int] = None,
                               event_fields=None) -> 'EventStream':
        from joule.api.event_stream import event_stream_get
        return await event_stream_get(self.session, stream, create,
                                      description=description,
                                      chunk_duration=chunk_duration,
                                      chunk_duration_us=chunk_duration_us,
                                      event_fields=event_fields)

    async def event_stream_move(self,
                                stream: 'EventStream | str | int',
                                folder: 'Folder | str | int') -> None:
        from joule.api.event_stream import event_stream_move
        return await event_stream_move(self.session, stream, folder)

    async def event_stream_update(self,
                                  stream: 'EventStream') -> None:
        from joule.api.event_stream import event_stream_update
        return await event_stream_update(self.session,
                                         stream)

    async def event_stream_delete(self,
                                  stream: 'EventStream | str | int') -> None:
        from joule.api.event_stream import event_stream_delete
        await event_stream_delete(self.session, stream)

    async def event_stream_create(self,
                                  stream: 'EventStream',
                                  folder: 'Folder | str | int') -> 'EventStream':
        from joule.api.event_stream import event_stream_create
        return await event_stream_create(self.session, stream, folder)

    async def event_stream_info(self,
                                stream: 'EventStream | str | int') -> 'EventStreamInfo':
        from joule.api.event_stream import event_stream_info
        return await event_stream_info(self.session, stream)

    async def event_stream_write(self,
                                 stream: 'EventStream | str | int',
                                 events: List['Event']) -> List['Event']:
        from joule.api.event_stream import event_stream_write
        return await event_stream_write(self.session, stream, events)

    def event_stream_read(self,
                                stream: 'EventStream | str | int',
                                start: Optional[int]=None,
                                end: Optional[int]=None,
                                json_filter: Optional[Dict[str, str]]=None,
                                include_on_going_events=True,
                                block_size: Optional[int] = None) -> List['Event']:
        from joule.api.event_stream import event_stream_read
        return event_stream_read(self.session, stream=stream,
                                       start_time=start, end_time=end,
                                       json_filter=json_filter,
                                       include_on_going_events=include_on_going_events,
                                       block_size=block_size)

    async def event_stream_read_list(self,
                                stream: 'EventStream | str | int',
                                start: Optional[int]=None,
                                end: Optional[int]=None,
                                json_filter: Optional[Dict[str, str]]=None,
                                include_on_going_events=True,
                                limit=10_000) -> List['Event']:
        from joule.api.event_stream import event_stream_read_list
        return await event_stream_read_list(self.session, stream=stream,
                                       start_time=start, end_time=end,
                                       json_filter=json_filter,
                                       include_on_going_events=include_on_going_events,
                                       limit=limit)
    
    async def event_stream_count(self,
                                 stream: 'EventStream | str | int',
                                 start: Optional[int] = None,
                                 end: Optional[int] = None,
                                 json_filter: Optional[Dict[str, str]] = None,
                                 include_on_going_events=True) -> int:
        from joule.api.event_stream import event_stream_count
        return await event_stream_count(self.session, stream, start, end, json_filter, include_on_going_events)

    async def event_stream_remove(self,
                                  stream: 'EventStream | str | int',
                                  start: Optional[int] = None,
                                  end: Optional[int] = None,
                                  json_filter=None) -> None:
        from joule.api.event_stream import event_stream_remove
        return await event_stream_remove(self.session, stream, start, end, json_filter)

    # Data actions

    async def data_read(self,
                        stream: 'DataStream | str | int',
                        start: Optional[int] = None,
                        end: Optional[int] = None,
                        max_rows: Optional[int] = None) -> 'Pipe':
        from joule.api.data import data_read
        return await data_read(self.session, stream, start, end,
                               max_rows)

    async def data_read_array(self,
                              stream: 'DataStream | str | int',
                              start: Optional[int] = None,
                              end: Optional[int] = None,
                              max_rows: int = 10000,
                              flatten: bool = False
                              ):
        from joule.api.data import data_read_array
        return await data_read_array(self.session, stream, start, end, max_rows, flatten)

    async def data_subscribe(self,
                             stream: 'DataStream | str | int') -> 'Pipe':
        from joule.api.data import data_subscribe
        return await data_subscribe(self.session, stream)

    async def data_intervals(self,
                             stream: 'DataStream | str | int',
                             start: Optional[int] = None,
                             end: Optional[int] = None) -> List:
        from joule.api.data import data_intervals
        return await data_intervals(self.session, stream, start, end)

    async def data_consolidate(self,
                               stream: 'DataStream | str | int',
                               max_gap: int,
                               start: Optional[int] = None,
                               end: Optional[int] = None) -> List:
        from joule.api.data import data_consolidate
        return await data_consolidate(self.session, stream, start, end, max_gap)

    async def data_drop_decimations(self,
                                    stream: 'DataStream | str | int'):
        from joule.api.data import data_drop_decimations
        return await data_drop_decimations(self.session, stream)

    async def data_decimate(self,
                            stream: 'DataStream | str | int'):
        from joule.api.data import data_decimate
        return await data_decimate(self.session, stream)

    async def data_write(self, stream: 'DataStream | str | int',
                         start: Optional[int] = None,
                         end: Optional[int] = None,
                         merge_gap:int = 0) -> 'Pipe':
        from joule.api.data import data_write
        return await data_write(self.session, stream, start, end, merge_gap=merge_gap)

    async def data_delete(self, stream: 'DataStream | str | int',
                          start: Optional[int] = None,
                          end: Optional[int] = None) -> None:
        from joule.api.data import data_delete
        return await data_delete(self.session, stream, start, end)

    # Module actions

    async def module_list(self,
                          statistics: bool = False) -> List['Module']:
        from joule.api.module import module_list
        return await module_list(self.session, statistics)

    async def module_get(self,
                         module: 'Module | str | int',
                         statistics: bool = False) -> 'Module':
        from joule.api.module import module_get
        return await module_get(self.session, module, statistics)

    async def module_logs(self,
                          module: 'Module | str | int') -> List[str]:
        from joule.api.module import module_logs
        return await module_logs(self.session, module)

    # Proxy actions

    async def proxy_list(self) -> List['Proxy']:
        from joule.api.proxy import proxy_list
        return await proxy_list(self.session)

    async def proxy_get(self,
                        proxy: 'Proxy | str | int') -> 'Proxy':
        from joule.api.proxy import proxy_get
        return await proxy_get(self.session, proxy)

    # Master actions

    async def master_list(self) -> List['Master']:
        from joule.api.master import master_list
        return await master_list(self.session)

    async def master_delete(self, master_type: str, name: str) -> None:
        from joule.api.master import master_delete
        return await master_delete(self.session, master_type, name)

    async def master_add(self, master_type, identifier,
                         lumen_parameters: Optional[Dict] = None,
                         api_key=None) -> 'Master':
        from joule.api.master import master_add
        return await master_add(self.session, master_type, identifier, lumen_parameters, api_key)

    async def follower_list(self) -> List['BaseNode']:
        from joule.api.follower import follower_list
        return await follower_list(self.session)
        
    async def follower_delete(self, node: 'BaseNode | str'):
        from joule.api.follower import follower_delete
        return await follower_delete(self.session, node)
        
    # Annotation actions
    async def annotation_get(self, stream: 'DataStream | str | int',
                             start: Optional[int] = None,
                             end: Optional[int] = None) -> List['Annotation']:
        from joule.api.annotation import annotation_get
        return await annotation_get(self.session, stream, start, end)

    async def annotation_create(self,
                                annotation: 'Annotation',
                                stream: 'DataStream | str | int', ) -> 'Annotation':
        from joule.api.annotation import annotation_create
        return await annotation_create(self.session,
                                       annotation, stream)

    async def annotation_delete(self,
                                annotation: 'Annotation | int'):
        from joule.api.annotation import annotation_delete
        return await annotation_delete(self.session, annotation)

    async def annotation_update(self,
                                annotation: 'Annotation'):
        from joule.api.annotation import annotation_update
        return await annotation_update(self.session, annotation)

    async def annotation_info(self,
                              stream: 'DataStream | str | int', ) -> 'AnnotationInfo':
        from joule.api.annotation import annotation_info
        return await annotation_info(self.session, stream)

    # Database actions
    async def db_connect(self) -> 'sqlalchemy.engine.Engine':
        from joule.api.db import db_connect
        return await db_connect(self.session)

    async def db_connection_info(self) -> 'ConnectionInfo':
        from joule.api.db import db_connection_info
        return await db_connection_info(self.session)



import unittest

from joule.utilities import misc


class TestMiscUtilities(unittest.TestCase):

    def test_yesno(self):
        self.assertTrue(misc.yesno("yes"))
        self.assertFalse(misc.yesno("no"))
        for val in ["badval", "", None]:
            with self.assertRaises(ValueError):
                misc.yesno(val)



import unittest

from joule.models import element
from joule.models.element import Element, from_json
from tests import helpers
from joule.errors import ConfigurationError

class TestElement(unittest.TestCase):

    def setUp(self):
        config = helpers.parse_configs(
            """[Element1]
                 name = test
                 plottable = no
                 # None uses the default value
                 offset = None
                 default_min = -10.5
            """)
        self.base_config = config['Element1']

    def test_parses_base_config(self):
        e = element.from_config(self.base_config)
        self.assertEqual(e.name, "test")
        self.assertFalse(e.plottable)
        self.assertEqual(e.display_type, Element.DISPLAYTYPE.CONTINUOUS)  # default value
        self.assertEqual(e.offset, 0.0)  # default value
        self.assertEqual(e.scale_factor, 1.0)  # default value
        self.assertEqual(e.default_min, -10.5)

    def test_merges_config(self):
        e1 = Element(name="test",
                    units="W",
                    plottable=True,
                    offset=0,
                    scale_factor=1,
                    display_type=Element.DISPLAYTYPE.EVENT,
                    default_min=None,
                    default_max=None)
        e2 = from_json(e1.to_json())
        e1.index=1
        e2.index=1
        # nothing to merge
        self.assertFalse(e1.merge_configs(e2))
        # merges new configs
        e2.name='new name'
        self.assertTrue(e1.merge_configs(e2))
        self.assertEqual(e1.name, 'new name')
        self.assertEqual(e1.units, 'W')
        # cannot merge different indices
        e2.index=2
        with self.assertRaisesRegex(ConfigurationError, 'cannot merge'):
            e1.merge_configs(e2)

       

    def test_creates_default_display_type(self):
        e = Element(name="test",
                    units="W",
                    plottable=True,
                    offset=0,
                    scale_factor=1,
                    display_type=None,
                    default_min=None,
                    default_max=None)
        element_json = e.to_json()
        self.assertEqual('CONTINUOUS', element_json['display_type'])

    def test_equality(self):
        e1 = Element(name="test",
                     units="W",
                     plottable=True,
                     offset=0,
                     scale_factor=1,
                     display_type=Element.DISPLAYTYPE.EVENT,
                     default_min=None,
                     default_max=None)
        e2 = Element(name="test",
                     units="W",
                     plottable=True,
                     offset=0,
                     scale_factor=1,
                     display_type=Element.DISPLAYTYPE.EVENT,
                     default_min=None,
                     default_max=None)
        self.assertEqual(e1, e2)
        e2.name = "different"
        self.assertNotEqual(e1, e2)
        self.assertNotEqual(e1,'not equal')

    def test_updates(self):
        e = Element(name="test",
                    units="W",
                    plottable=True,
                    offset=0,
                    scale_factor=1,
                    display_type=Element.DISPLAYTYPE.EVENT,
                    default_min=None,
                    default_max=None)
        e.update_attributes({
            "name": "new name",
            "units": "nW",
            "plottable": False,
            "offset": -1,
            "scale_factor": 10,
            "display_type": "discrete",
            "default_min": 8,
            "default_max": 100
        })
        self.assertEqual("new name", e.name)
        self.assertEqual("nW", e.units)
        self.assertFalse(e.plottable)
        self.assertEqual(-1, e.offset)
        self.assertEqual(10, e.scale_factor)
        self.assertEqual(Element.DISPLAYTYPE.DISCRETE, e.display_type)
        self.assertEqual(8, e.default_min)
        self.assertEqual(100, e.default_max)



# Shared validator functions for the backend and API tools
from typing import Dict
import json
from joule.errors import ConfigurationError

def validate_event_fields(fields: Dict[str, str]) -> Dict[str, str]:
    """ 
    make sure values are either string, numeric, or category
    category has a JSON formatted list of categories after a ':' for example:
        category:["cat1","cat2","cat3"]
    """
    if not isinstance(fields, dict):
        raise ConfigurationError(f"event_fields must be a dictionary, not a {type(fields).__name__}")
    for dt_string in fields.values():
        data_type = dt_string.split(':')[0]
        if data_type not in ['string', 'numeric','category']:
            raise ConfigurationError("invalid value in event_fields, must be [numeric|string|category:cat1,cat2,...]")
        if data_type == 'category':
            try:
                # make sure there is a : and a JSON parseable list of strings after it
                if ':' not in dt_string:
                    raise ValueError()
                categories = json.loads(dt_string.split(':')[1])
                # make sure categories is a list of strings
                if not isinstance(categories, list):
                    raise ValueError()
                if len(categories) == 0:
                    raise ValueError()
                for category in categories:
                    if not isinstance(category, str):
                        raise ValueError()
            except (json.JSONDecodeError, ValueError):
                raise ConfigurationError("invalid value in event_fields, category type must a have JSON list of categories after a ':'")
    return fields



from aiohttp.test_utils import AioHTTPTestCase, unittest_run_loop
from aiohttp import web
import aiohttp
import numpy as np
from sqlalchemy.orm import Session
import asyncio
from joule.models import EventStream, Folder
import joule.controllers
from tests.controllers.helpers import create_db, MockSupervisor, MockEventStore
from joule.api.event_stream import EventStream as ApiEventStream
from joule.models.folder import find as find_folder
from joule.api.event import Event
from tests import helpers
from joule import app_keys
from joule.constants import EndPoints, ApiErrorMessages

import testing.postgresql
psql_key = web.AppKey("psql", testing.postgresql.Postgresql)

class TestEventController(AioHTTPTestCase):

    async def tearDownAsync(self):
        self.app[app_keys.db].close()
        await asyncio.to_thread(self.app[psql_key].stop)
        await self.client.close()

    async def get_application(self):
        app = web.Application()
        app.add_routes(joule.controllers.routes)
        stream1 = ApiEventStream(name="test",description="test stream",keep_us=10e6, 
                              event_fields={"field1": "string",
                                           "field2": 'category:["1","2","3"]'})
        stream2 = ApiEventStream(name="test2",description="test stream",keep_us=10e6, 
                              event_fields={"field1": "string",
                                           "field2": 'category:["1","2","3"]'})
        app[app_keys.db], app[psql_key] = await asyncio.to_thread(lambda: create_db([],[stream1, stream2]))
        self.test_stream_id = app[app_keys.db].query(EventStream).filter_by(name="test").one().id
        self.test_stream = stream1 # stream2 is to test for name collisions
        self.test_stream.id = self.test_stream_id # set the id to match the database
        app[app_keys.event_store] = MockEventStore()
        self.supervisor = MockSupervisor()
        app[app_keys.supervisor] = self.supervisor
        return app
    
    async def test_create_event_stream_errors(self):
        store: MockEventStore = self.app[app_keys.event_store]
        new_stream = ApiEventStream(name="test",description="test stream",keep_us=10e6, 
                                 event_fields={"field1": "string",
                                               "field2": 'category:["1","2","3"]'})
        ### must send JSON data
        resp: aiohttp.ClientResponse = await \
            self.client.post(EndPoints.event, params={"dest_path": "/folder1", "stream": "not json"})
        self.assertEqual(resp.status, 400)
        msg = await resp.text()
        self.assertIn('content-type', msg)
        self.assertFalse(store.create_called)

        ### must send dest_path
        resp: aiohttp.ClientResponse = await \
            self.client.post(EndPoints.event, json={"stream": new_stream.to_json()})
        self.assertEqual(resp.status, 400)
        msg = await resp.text()
        self.assertIn('destination', msg)
        self.assertFalse(store.create_called)

        ### must send stream
        resp: aiohttp.ClientResponse = await \
            self.client.post(EndPoints.event, json={"dest_path": "/folder1"})
        self.assertEqual(resp.status, 400)
        msg = await resp.text()
        self.assertIn('stream', msg)
        self.assertFalse(store.create_called)

        ### destination must be valid (missing a /)
        resp: aiohttp.ClientResponse = await \
            self.client.post(EndPoints.event, json={"dest_path": "folder1", "stream": new_stream.to_json()})
        self.assertEqual(resp.status, 400)
        msg = await resp.text()
        self.assertIn('invalid path', msg)
        self.assertFalse(store.create_called)

        ### stream JSON must be valid
        resp: aiohttp.ClientResponse = await \
            self.client.post(EndPoints.event, json={"dest_path": "/folder1", "stream": "Not JSON :("})
        self.assertEqual(resp.status, 400)
        msg = await resp.text()
        self.assertIn('JSON', msg)
        self.assertFalse(store.create_called)

        ### stream JSON must have all the correct fields
        stream_json = new_stream.to_json()
        stream_json.pop('name')
        resp: aiohttp.ClientResponse = await \
            self.client.post(EndPoints.event, json={"dest_path": "/folder1", "stream": stream_json})
        self.assertEqual(resp.status, 400)
        msg = await resp.text()
        self.assertIn('missing stream attribute', msg)
        self.assertFalse(store.create_called)

    async def test_delete_event_stream_errors(self):
        ### must send path or id
        prev_update = self.app[app_keys.db].query(Folder).filter_by(name="events").one().updated_at
        resp: aiohttp.ClientResponse = await \
            self.client.delete(EndPoints.event, params={})
        self.assertEqual(resp.status, 400)
        msg = await resp.text()
        self.assertIn('path', msg)
        # make sure the parent folder timestamp is not updated
        updated_at = self.app[app_keys.db].query(Folder).filter_by(name="events").one().updated_at
        self.assertEqual(updated_at, prev_update)

        ## path must exist
        resp: aiohttp.ClientResponse = await \
            self.client.delete(EndPoints.event, params={"path": "/does/not/exist"})
        self.assertEqual(resp.status, 404)
        msg = await resp.text()
        self.assertIn('does not exist', msg)
        # make sure the parent folder timestamp is not updated
        updated_at = self.app[app_keys.db].query(Folder).filter_by(name="events").one().updated_at
        self.assertEqual(updated_at, prev_update)
    
    async def test_get_event_stream_info_errors(self):
        ### must send id
        resp: aiohttp.ClientResponse = await \
            self.client.get(EndPoints.event, params={})
        self.assertEqual(resp.status, 400)
        msg = await resp.text()
        self.assertIn('id', msg)

        ### id must exist
        resp: aiohttp.ClientResponse = await \
            self.client.get(EndPoints.event, params={"id": 999})
        self.assertEqual(resp.status, 404)
        msg = await resp.text()
        self.assertIn('does not exist', msg)

    async def test_move_event_stream_errors(self):
        ### must send id
        resp: aiohttp.ClientResponse = await \
            self.client.put(EndPoints.event_move, json={})
        self.assertEqual(resp.status, 400)
        msg = await resp.text()
        self.assertIn('id', msg)

        ### id must exist
        resp: aiohttp.ClientResponse = await \
            self.client.put(EndPoints.event_move, json={"src_id": 999})
        self.assertEqual(resp.status, 404)
        msg = await resp.text()
        self.assertIn('does not exist', msg)

        ### must send dest_path
        resp: aiohttp.ClientResponse = await \
            self.client.put(EndPoints.event_move, json={"src_id": 1})
        self.assertEqual(resp.status, 400)
        msg = await resp.text()
        self.assertIn('destination', msg)

        ### dest_path must be valid
        resp: aiohttp.ClientResponse = await \
            self.client.put(EndPoints.event_move, json={"src_id": 1, "dest_path": "not/valid"})
        self.assertEqual(resp.status, 400)
        
        ### name must be unique in destination
        duplicate_stream = ApiEventStream(name="test",description="test stream")
        resp: aiohttp.ClientResponse = await \
            self.client.post(EndPoints.event, json={"dest_path": "/other/folder", "stream": duplicate_stream.to_json()})
        self.assertEqual(resp.status, 200)
        resp: aiohttp.ClientResponse = await \
            self.client.put(EndPoints.event_move, json={"src_path": "/events/test", "dest_path": "/other/folder"})
        self.assertEqual(resp.status, 400)
        msg = await resp.text()
        self.assertIn('same name', msg)
        # make sure the stream didn't move, the parent_id should be the same
        actual_parent_id = self.app[app_keys.db].query(EventStream).filter_by(id=self.test_stream_id).one().folder_id
        self.assertEqual(actual_parent_id, find_folder("/events", self.app[app_keys.db]).id)

    async def test_event_stream_update_errors(self):
        db = self.app[app_keys.db]
        stream = db.query(EventStream).filter_by(id=self.test_stream_id).one()
        prev_update = stream.updated_at
        
        ### must send JSON data for stream
        resp: aiohttp.ClientResponse = await \
            self.client.put(EndPoints.event, json={"id": self.test_stream_id})
        self.assertEqual(resp.status, 400)
        # database is not updated
        stream = db.query(EventStream).filter_by(id=self.test_stream_id).one()
        self.assertEqual(stream.updated_at, prev_update)

        ### JSON data must be valid
        resp: aiohttp.ClientResponse = await \
            self.client.put(EndPoints.event, headers={'Content-Type':'application/json'}, data='not json')
        self.assertEqual(resp.status, 400)
        # database is not updated
        stream = db.query(EventStream).filter_by(id=self.test_stream_id).one()
        self.assertEqual(stream.updated_at, prev_update)
        
        ### must send id
        resp: aiohttp.ClientResponse = await \
            self.client.put(EndPoints.event, json={"stream": self.test_stream.to_json()})
        self.assertEqual(resp.status, 400)
        # database is not updated
        stream = db.query(EventStream).filter_by(id=self.test_stream_id).one()
        self.assertEqual(stream.updated_at, prev_update)

        ### id must be valid
        resp: aiohttp.ClientResponse = await \
            self.client.put(EndPoints.event, json={"id":999, "stream": self.test_stream.to_json()})
        self.assertEqual(resp.status, 404)
        # database is not updated
        stream = db.query(EventStream).filter_by(id=self.test_stream_id).one()
        self.assertEqual(stream.updated_at, prev_update)

        ### stream JSON must be valid
        resp: aiohttp.ClientResponse = await \
            self.client.put(EndPoints.event, json={"id": self.test_stream_id, "stream": "not json"})
        self.assertEqual(resp.status, 400)
        # database is not updated
        stream = db.query(EventStream).filter_by(id=self.test_stream_id).one()
        self.assertEqual(stream.updated_at, prev_update)

        ### name must be unique in the folder
        self.test_stream.name="test2" # conflict
        resp: aiohttp.ClientResponse = await \
            self.client.put(EndPoints.event, json={"id":self.test_stream_id, "stream": self.test_stream.to_json()})
        self.assertEqual(resp.status, 400)
        msg = await resp.text()
        self.assertIn('same name', msg)

    async def test_event_stream_write_errors(self):
        event_store = self.app[app_keys.event_store]
        ### must specify an id or path
        events = [Event(start_time=i,end_time=i+1000,content={"field1": i}, 
                        event_stream_id=self.test_stream.id, id=i) for i in range(0,10000,1000)]
        resp: aiohttp.ClientResponse = await \
            self.client.post(EndPoints.event_data, json={"events": [e.to_json(destination_node_uuid='test') for e in events]})
        self.assertEqual(resp.status, 400)
        self.assertFalse(event_store.upsert_called)
        msg = await resp.text()
        self.assertIn(ApiErrorMessages.specify_id_or_path, msg)

        ### stream must exist (valid path)
        resp: aiohttp.ClientResponse = await \
            self.client.post(EndPoints.event_data, json={"events": [e.to_json(destination_node_uuid='test') for e in events], "path": "/does/not/exist"})
        self.assertEqual(resp.status, 404)
        self.assertFalse(event_store.upsert_called)
        msg = await resp.text()
        self.assertIn(ApiErrorMessages.stream_does_not_exist, msg)

        ### must send events
        resp: aiohttp.ClientResponse = await \
            self.client.post(EndPoints.event_data, json={"path": "/events/test"})
        self.assertEqual(resp.status, 400)
        self.assertFalse(event_store.upsert_called)
        msg = await resp.text()
        self.assertIn('events', msg)

        ### events must have an event_stream_id
        events_json = [e.to_json(destination_node_uuid='test') for e in events]
        events_json[0].pop('event_stream_id')
        resp: aiohttp.ClientResponse = await \
            self.client.post(EndPoints.event_data, json={"events": events_json, "path": "/events/test"})
        self.assertEqual(resp.status, 400)
        self.assertFalse(event_store.upsert_called)
        msg = await resp.text()
        self.assertIn('event_stream_id', msg)

    async def test_event_stream_count_errors(self):
        event_store = self.app[app_keys.event_store]
        ### must specify an id or path
        resp: aiohttp.ClientResponse = await \
            self.client.get(EndPoints.event_data_count, params={"start": 0, "end": 10000})
        self.assertEqual(resp.status, 400)
        self.assertFalse(event_store.count_called)
        msg = await resp.text()
        self.assertIn(ApiErrorMessages.specify_id_or_path, msg)

        ### stream must exist (valid path)
        resp: aiohttp.ClientResponse = await \
            self.client.get(EndPoints.event_data_count, params={"path": "/does/not/exist"})
        self.assertEqual(resp.status, 404)
        self.assertFalse(event_store.count_called)
        msg = await resp.text()
        self.assertIn(ApiErrorMessages.stream_does_not_exist, msg)

        ### parameters must have valid datatype (int)
        resp: aiohttp.ClientResponse = await \
            self.client.get(EndPoints.event_data_count, params={"path": "/events/test", "start": "not an int", "end": 10000})
        self.assertEqual(resp.status, 400)
        self.assertFalse(event_store.count_called)
        msg = await resp.text()
        self.assertIn(ApiErrorMessages.f_parameter_must_be_an_int.format(parameter='start'), msg)

        ### start must be before end
        resp: aiohttp.ClientResponse = await \
            self.client.get(EndPoints.event_data_count, params={"path": "/events/test", "start": 10000, "end": 0})
        self.assertEqual(resp.status, 400)
        self.assertFalse(event_store.count_called)
        msg = await resp.text()
        self.assertIn(ApiErrorMessages.start_must_be_before_end, msg)

        ### invalid filter parameter
        resp: aiohttp.ClientResponse = await \
            self.client.get(EndPoints.event_data_count, params={"path": "/events/test", "start": 0, "end": 10000, "filter": "not valid"})   
        self.assertEqual(resp.status, 400)
        self.assertFalse(event_store.count_called)
        msg = await resp.text()
        self.assertIn(ApiErrorMessages.invalid_filter_parameter, msg)

    async def test_event_stream_read_errors(self):
        event_store = self.app[app_keys.event_store]
        ### must specify an id or path
        resp: aiohttp.ClientResponse = await \
            self.client.get(EndPoints.event_data, params={"start": 0, "end": 10000})
        self.assertEqual(resp.status, 400)
        self.assertFalse(event_store.count_called)
        self.assertFalse(event_store.extract_called)
        msg = await resp.text()
        self.assertIn(ApiErrorMessages.specify_id_or_path, msg)

        ### stream must exist (valid path)
        resp: aiohttp.ClientResponse = await \
            self.client.get(EndPoints.event_data, params={"path": "/does/not/exist"})
        self.assertEqual(resp.status, 404)
        self.assertFalse(event_store.count_called)
        self.assertFalse(event_store.extract_called)
        msg = await resp.text()
        self.assertIn(ApiErrorMessages.stream_does_not_exist, msg)

        ### limit must be provided
        resp: aiohttp.ClientResponse = await \
            self.client.get(EndPoints.event_data, params={"path": "/events/test"})
        self.assertEqual(resp.status, 400)
        self.assertFalse(event_store.count_called)
        self.assertFalse(event_store.extract_called)
        msg = await resp.text()
        self.assertIn("limit", msg)

        ### limit must be > 0
        resp: aiohttp.ClientResponse = await \
            self.client.get(EndPoints.event_data, params={"path": "/events/test", "limit": -5})
        self.assertEqual(resp.status, 400)
        self.assertFalse(event_store.count_called)
        self.assertFalse(event_store.extract_called)
        msg = await resp.text()
        self.assertIn("limit", msg)

    async def test_event_stream_remove_errors(self):
        event_store = self.app[app_keys.event_store]
        ### must specify an id or path
        resp: aiohttp.ClientResponse = await \
            self.client.delete(EndPoints.event_data, params={"start": 0, "end": 10000})
        self.assertEqual(resp.status, 400)
        self.assertFalse(event_store.remove_called)
        msg = await resp.text()
        self.assertIn(ApiErrorMessages.specify_id_or_path, msg)

        ### stream must exist (valid path)
        resp: aiohttp.ClientResponse = await \
            self.client.delete(EndPoints.event_data, params={"path": "/does/not/exist"})
        self.assertEqual(resp.status, 404)
        self.assertFalse(event_store.remove_called)
        msg = await resp.text()
        self.assertIn(ApiErrorMessages.stream_does_not_exist, msg)

        ### parameters must have valid datatype (int)
        resp: aiohttp.ClientResponse = await \
            self.client.delete(EndPoints.event_data, params={"path": "/events/test", "start": "not an int", "end": 10000})
        self.assertEqual(resp.status, 400)
        self.assertFalse(event_store.remove_called)
        msg = await resp.text()
        self.assertIn(ApiErrorMessages.f_parameter_must_be_an_int.format(parameter='start'), msg)

        ### start must be before end
        resp: aiohttp.ClientResponse = await \
            self.client.delete(EndPoints.event_data, params={"path": "/events/test", "start": 10000, "end": 0})
        self.assertEqual(resp.status, 400)
        self.assertFalse(event_store.remove_called)
        msg = await resp.text()
        self.assertIn(ApiErrorMessages.start_must_be_before_end, msg)


import unittest
from joule import api, errors


class TestStreamMethods(unittest.IsolatedAsyncioTestCase):

    async def asyncSetUp(self):
        self.node = api.get_node()
        """
        └── test
            ├── f1
            │   ├── f1a
            │   │   ├── s1a1
            │   │   └── s1a2
            │   └── s1a
            └── f2
                └── f2a
                    └── s2a1a
        """
        stream = api.DataStream(name="s1a1", elements=[api.Element(name="e1")])
        await self.node.data_stream_create(stream, "/test/f1/f1a")
        stream.name = "s1a2"
        await self.node.data_stream_create(stream, "/test/f1/f1a")
        stream.name = "s1a"
        await self.node.data_stream_create(stream, "/test/f1")
        stream.name = "s2a1a"
        await self.node.data_stream_create(stream, "/test/f2/f2a")

    async def asyncTearDown(self):
        await self.node.folder_delete("/test", recursive=True)
        await self.node.close()

    async def test_stream_move_to_existing_folder(self):
        await self.node.data_stream_move("/test/f1/f1a/s1a1", "/test/f2/f2a")
        stream = await self.node.data_stream_get("/test/f2/f2a/s1a1")
        self.assertEqual("s1a1", stream.name)

    async def test_stream_move_to_new_folder(self):
        await self.node.data_stream_move("/test/f1/f1a/s1a1", "/test/f2/new")
        stream = await self.node.data_stream_get("/test/f2/new/s1a1")
        self.assertEqual("s1a1", stream.name)

    async def test_move_errors(self):
        # stream does not exist
        with self.assertRaises(errors.ApiError):
            await self.node.data_stream_move("/test/f1/f1a/bad", "/test/f2/new")

        # name conflict
        stream = api.DataStream(name="s1a1", elements=[api.Element(name="e1")])
        await self.node.data_stream_create(stream, "/test/f2")

        with self.assertRaises(errors.ApiError):
            await self.node.data_stream_move(stream, "/test/f1/f1a")

        # locked stream
        with self.assertRaises(errors.ApiError):
            await self.node.data_stream_move("/live/base", "/test/f2/new")

    async def test_stream_update(self):
        stream = await self.node.data_stream_get("/test/f1/f1a/s1a1")
        stream.name = "new name"
        stream.description = "new description"
        stream.elements[0].name = "new element name"
        await self.node.data_stream_update(stream)
        stream = await self.node.data_stream_get(stream)
        self.assertEqual(stream.name, "new name")
        self.assertEqual(stream.description, "new description")
        self.assertEqual(stream.elements[0].name, "new element name")

    async def testStreamUpdateErrors(self):
        # locked stream
        stream = await self.node.data_stream_get("/live/base")
        with self.assertRaises(errors.ApiError):
            stream.name = "invalid"
            await self.node.data_stream_update(stream)

        # cannot change number of elements
        stream = await self.node.data_stream_get("/test/f1/f1a/s1a1")
        with self.assertRaises(errors.ApiError):
            stream.elements.append(api.Element(name="invalid"))
            await self.node.data_stream_update(stream)

        # name conflict
        stream = await self.node.data_stream_get("/test/f1/f1a/s1a1")
        with self.assertRaises(errors.ApiError):
            stream.name = "s1a2"
            await self.node.data_stream_update(stream)

    async def testStreamCreate(self):
        # this is done in the setUp calls already
        pass

    async def testStreamCreateErrors(self):
        # name conflict
        stream = api.DataStream(name="s1a1", elements=[api.Element(name="e1")])
        with self.assertRaises(errors.ApiError):
            await self.node.data_stream_create(stream, "/test/f1/f1a")

        # missing elements
        stream = api.DataStream(name="no elements", elements=[])
        with self.assertRaises(errors.ApiError):
            await self.node.data_stream_create(stream, "/test/f1/f1a")

        # invalid values
        stream = api.DataStream(name="invalid datatype", datatype="invalid",
                                elements=[api.Element(name="e1")])
        with self.assertRaises(errors.ApiError):
            await self.node.data_stream_create(stream, "/test/f1/f1a")

    async def testStreamDelete(self):
        await self.node.data_stream_delete("/test/f1/f1a/s1a1")
        with self.assertRaises(errors.ApiError):
            await self.node.data_stream_get("/test/f1/f1a/s1a1")

    async def testStreamDeleteErrors(self):
        # missing stream
        with self.assertRaises(errors.ApiError):
            await self.node.data_stream_delete("/test/f1/f1a/bad")

        # locked stream
        with self.assertRaises(errors.ApiError):
            await self.node.data_stream_delete("/live/plus1")
        stream = await self.node.data_stream_get("/live/plus1")
        self.assertEqual(stream.name, "plus1")

    async def testStreamInfo(self):
        info = await self.node.data_stream_info("/live/base")
        self.assertGreater(info.total_time, 0)

        info = await self.node.data_stream_info("/archive/data2")
        self.assertEqual(info.total_time, 0)

    async def testStreamInfoErrors(self):
        # missing stream
        with self.assertRaises(errors.ApiError):
            await self.node.data_stream_info("/test/bad")



from typing import List, TYPE_CHECKING, Dict, Union

from .session import BaseSession
from joule.constants import EndPoints
from .node import TcpNode
if TYPE_CHECKING:
    from .node import BaseNode

async def follower_list(session: BaseSession) -> List['BaseNode']:
    resp = await session.get(EndPoints.followers)
    return [TcpNode(item['name'], item['location'], item['key'], session.cafile) for item in resp]

async def follower_delete(session: BaseSession, node: 'BaseNode | str') -> None:
    if type(node) is not str:
        name = node.name
    else:
        name = node
    data = {"name": name}
    await session.delete(EndPoints.follower, params=data)



from click.testing import CliRunner
import os
from aiohttp.test_utils import unused_port
import warnings
from ..fake_joule import FakeJoule, FakeJouleTestCase
from joule.cli import main

MODULE_LOGS = os.path.join(os.path.dirname(__file__), 'logs.json')
warnings.simplefilter('always')


class TestModuleLogs(FakeJouleTestCase):

    def test_print_module_logs(self):
        server = FakeJoule()
        with open(MODULE_LOGS, 'r') as f:
            server.response = f.read()
        self.start_server(server)
        runner = CliRunner()
        result = runner.invoke(main, ['module', 'logs', 'my_module'])
        self.assertEqual(result.exit_code, 0)
        output = result.output
        self.assertGreater(len(output), 0)
        self.assertIn("starting", output)
        self.stop_server()

    def test_when_server_returns_invalid_data(self):
        server = FakeJoule()
        server.response = "notjson"
        self.start_server(server)
        runner = CliRunner()
        result = runner.invoke(main, ['module', 'logs', 'my_module'])
        self.assertIn('Error', result.output)
        self.assertEqual(result.exit_code, 1)
        self.stop_server()

    def test_when_server_returns_error_code(self):
        server = FakeJoule()
        error_msg = "test error"
        error_code = 500
        server.response = error_msg
        server.http_code = error_code
        self.start_server(server)
        runner = CliRunner()
        result = runner.invoke(main, ['module', 'logs', 'my_module'])
        self.assertIn('%d' % error_code, result.output)
        self.assertIn(error_msg, result.output)
        self.assertEqual(result.exit_code, 1)
        self.stop_server()



import unittest

from joule.models import Element
from joule.models import element
from joule.errors import ConfigurationError
from tests import helpers


class TestElementErrors(unittest.TestCase):

    def setUp(self):
        config = helpers.parse_configs(
            """[Element1]
                 name = test
                 plottable = no
                 offset = 5.2
            """)
        self.base_config = config['Element1']

    def test_must_have_name(self):
        self.base_config['name'] = ""
        with self.assertRaisesRegex(ConfigurationError, "name"):
            element.from_config(self.base_config)

    def test_sensible_bounds(self):
        """default_min<default_max or both == 0 for autoscale"""
        self.base_config['default_min'] = '10'
        self.base_config['default_max'] = '-10'
        with self.assertRaisesRegex(ConfigurationError, "default_min"):
            element.from_config(self.base_config)

    def test_errors_on_bad_offsets(self):
        bad_offsets = ['a', '*', '0y']
        self.evaluate_bad_values("offset", bad_offsets)

    def test_errors_on_bad_scale_factor(self):
        bad_values = ['a', '*', '0y']
        self.evaluate_bad_values("scale_factor", bad_values)

    def test_errors_on_bad_default_min(self):
        bad_values = ['a', '*', '0y']
        self.evaluate_bad_values("default_min", bad_values)

    def test_errors_on_bad_default_max(self):
        bad_values = ['a', '*', '0y']
        self.evaluate_bad_values("default_max", bad_values)

    def test_errors_on_bad_bool(self):
        bad_values = ['asdf', 'not_valid']
        self.evaluate_bad_values("plottable", bad_values)

    def test_errors_on_bad_display_type(self):
        bad_values = ['lollipops', '', 'circles']
        self.evaluate_bad_values("display_type", bad_values)

    def evaluate_bad_values(self, setting_name, bad_settings):
        for setting in bad_settings:
            with self.subTest(setting=setting):
                self.base_config[setting_name] = setting
                with self.assertRaisesRegex(ConfigurationError, setting_name):
                    element.from_config(self.base_config)

    def test_errors_on_invalid_update(self):
        # value types must be correct
        e = Element(name="test")
        with self.assertRaises(ConfigurationError):
            e.update_attributes({
                "name": "new name",
                "default_min": 'invalid',
            })
        with self.assertRaises(ConfigurationError):
            e.update_attributes({
                "name": "new name",
                "offset": '',
            })
        # default_min < default_max
        with self.assertRaises(ConfigurationError) as error:
            e.update_attributes({
                "name": "new name",
                "default_min": 100,
                "default_max": 10
            })
        self.assertIn('default_min', str(error.exception))


import click

from joule.cli.config import Config, pass_config
from joule.cli.lazy_group import LazyGroup

@click.group(
    cls=LazyGroup,
    lazy_subcommands={
        "data": "joule.cli.data.data",
        "stream": "joule.cli.stream.streams",
        "module": "joule.cli.module.module",
        "folder": "joule.cli.folder.folders",
        "proxy": "joule.cli.proxy.proxies",
        "admin": "joule.cli.admin.admin",
        "master": "joule.cli.master.master",
        "follower": "joule.cli.follower.follower",
        "event": "joule.cli.event.events",
        "node": "joule.cli.node.node"
        }
)
@click.option('-n', '--node', default="", help="Joule Node name")
@click.version_option()
@pass_config
def main(config, node):
    # create a Node structure for the name
    # if node_name is given use it, otherwise go with the default
    config.set_node_name(node)




import logging
import tempfile
import os
import datetime
import unittest

from tests.helpers import DbTestCase
from joule.models import (DataStream, Folder,
                          Element, Module)
from joule.services import load_modules

logger = logging.getLogger('joule')


class TestConfigureModules(DbTestCase):

    def test_parses_configs(self):
        """e2e module configuration service test"""

        # /test/stream1:float32_3
        folder_test = Folder(name="test",
                             updated_at=datetime.datetime.now())
        stream1 = DataStream(name="stream1", keep_us=100,
                             datatype=DataStream.DATATYPE.FLOAT32,
                             updated_at=datetime.datetime.now())
        stream1.elements = [Element(name="e%d" % x, index=x, default_min=1) for x in range(3)]
        folder_test.data_streams.append(stream1)

        # /test/deeper/stream2: int16_2
        folder_deeper = Folder(name="deeper",
                               updated_at=datetime.datetime.now())
        stream2 = DataStream(name="stream2", datatype=DataStream.DATATYPE.INT16,
                             updated_at=datetime.datetime.now())
        stream2.elements = [Element(name="e%d" % x, index=x) for x in range(2)]
        folder_deeper.data_streams.append(stream2)
        folder_deeper.parent = folder_test

        root = Folder(name="root",
                      updated_at=datetime.datetime.now())
        root.children = [folder_test]
        self.db.add(root)

        self.db.commit()
        configs = [
            # writes to /test/stream1
            """
            [Main] 
              name = module1
              exec_cmd = runit.sh
            [Arguments]
              key = value
            # no inputs
            [Outputs]
              raw = /test/stream1
            """,
            # reads from /test/stream1, writes to /test/deeper/stream2 and /test/stream3
            """
            [Main]
              name = module2
              exec_cmd = runit2.sh
            [Inputs]
              source = /test/stream1:float32[e0,e1, e2]
            [Outputs]
              sink1 = /test/deeper/stream2
              sink2 = /test/stream3:int16[ x, y ]
            """,
            # ignored: unconfigured input
            """
            [Main]
              name = bad_module
              exec_cmd = runit3.sh
            [Inputs]
              source = /missing/stream
            # no outputs
            """,
            # ignored: mismatched stream config
            """
            [Main] 
              name = bad_module2
              exec_cmd = runit4.sh
            [Inputs]
              source = /test/stream3:int16[x,y,z]
            [Outputs]
            """,
        ]
        with tempfile.TemporaryDirectory() as conf_dir:
            # write the configs in 0.conf, 1.conf, ...
            i = 0
            for conf in configs:
                with open(os.path.join(conf_dir, "%d.conf" % i), 'w') as f:
                    f.write(conf)
                i += 1
            with self.assertLogs(logger=logger, level=logging.ERROR) as logs:
                modules = load_modules.run(conf_dir, self.db)
                output = ' '.join(logs.output)
                # log the missing stream configuration
                self.assertIn('/missing/stream', output)
                # log the incompatible stream configuration
                self.assertIn('different elements', output)
        # now check the database:
        # should have three streams
        self.assertEqual(self.db.query(DataStream).count(), 3)
        # and two modules
        self.assertEqual(len(modules), 2)
        # module1 should have no inputs and one output
        m1: Module = [m for m in modules if m.name == "module1"][0]
        self.assertEqual(len(m1.inputs), 0)
        self.assertEqual(len(m1.outputs), 1)
        self.assertEqual(m1.outputs["raw"], stream1)
        # module2 should have 1 input and 2 outputs
        m2: Module = [m for m in modules if m.name == "module2"][0]
        self.assertEqual(len(m2.inputs), 1)
        self.assertEqual(len(m2.outputs), 2)
        self.assertEqual(m2.inputs["source"], stream1)
        self.assertEqual(m2.outputs['sink1'], stream2)
        # sink2 goes to a new stream
        stream3 = self.db.query(DataStream).filter_by(name="stream3").one()
        self.assertEqual(m2.outputs['sink2'], stream3)



import unittest
import sqlalchemy
import numpy as np
from joule import api, errors
from joule.utilities import datetime_to_timestamp as dt2ts

class TestUserMethods(unittest.IsolatedAsyncioTestCase):

    async def asyncSetUp(self):
        self.node = api.get_node()

    async def asyncTearDown(self) -> None:
        await self.node.close()

    async def test_master_list(self):
        masters = await self.node.master_list()
        master_names = [m.name for m in masters]
        # e2e is created by admin authorize and auto_user is created by the user file
        # in main_node.py
        self.assertListEqual(master_names, ['e2e', 'auto_user'])


    